# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest

class TestPipHelperMethods(unittest.TestCase):

    def test_run_pip_is_present_in_cli_core(self):
        try:
            from azure.cli.core.extension.operations import _run_pip
        except ImportError:
            self.fail('dependency on cli core is broken, pip helper will need fix')


if __name__ == '__main__':
    unittest.main()


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from azext_devops.dev.common.uri import (uri_parse_instance_from_git_uri)


class TestUriMethods(unittest.TestCase):

    def test_uri_parse_instance_from_git_uri(self):
        # simple validation that we don't get an exception
        uri = "https://mseng@dev.azure.com/mseng/azuredevops/_git/azuredevops"
        result = uri_parse_instance_from_git_uri(uri)
        self.assertEqual(result, "https://dev.azure.com/mseng")

        uri = "https://mseng.visualstudio.com/AzureDevOps/_git/AzureDevOps"
        result = uri_parse_instance_from_git_uri(uri)
        self.assertEqual(result, "https://mseng.visualstudio.com/")

        uri = "https://mseng.visualstudio.com/"
        result = uri_parse_instance_from_git_uri(uri)
        self.assertEqual(result, "https://mseng.visualstudio.com/")


if __name__ == '__main__':
    unittest.main()


"""
Fixes coverage reports by inlining CSS files.
see https://github.com/Microsoft/azure-pipelines-tasks/issues/3027
Based on https://stackoverflow.com/a/23190429/7417402
"""
import os

import bs4

COVERAGE_REPORT_DIR = 'azure-devops/htmlcov/'
COVERAGE_REPORT = os.path.join(COVERAGE_REPORT_DIR, 'index.html')


def embed_css_in_html_file(html_file, css_dir):
    with open(html_file, 'r') as f:
        soup = bs4.BeautifulSoup(f.read(), "html.parser")

    stylesheets = soup.findAll("link", {"rel": "stylesheet"})
    for s in stylesheets:
        t = soup.new_tag('style')
        css_file = s["href"]
        print(f"found link to {css_file}")
        with open(os.path.join(css_dir, css_file), 'r') as f:
            c = bs4.element.NavigableString(f.read())
        t.insert(0, c)
        t['type'] = 'text/css'
        s.replaceWith(t)

    with open(html_file, 'w') as f:
        f.write(str(soup))


for file in os.listdir(COVERAGE_REPORT_DIR):
    if file.endswith(".html"):
        print(f"Embedding CSS in {file}")
        embed_css_in_html_file(os.path.join(COVERAGE_REPORT_DIR, file), COVERAGE_REPORT_DIR)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azext_devops.dev.common.git import is_git_alias_setup, setup_git_alias, clear_git_alias


def setup_git_aliases(local=False):
    """Configure aliases for Git (to enable commands like 'git pr list')
    :param local: Sets the alias in the local git config rather than the global config.
    :type local: bool
    """
    for key in git_aliases:
        setup_git_alias(key, git_aliases[key], local=local)


def clear_git_aliases(local=False):
    """
    :param local: Checks the alias in the local git config rather than the global config.
    :type local: bool
    """
    for key in git_aliases:
        if is_git_alias_setup(key, git_aliases[key], local=local):
            clear_git_alias(alias=key, local=local)


def are_git_aliases_setup(local=False):
    """
    :param local: Checks the alias in the local git config rather than the global config.
    :type local: bool
    """
    for key in git_aliases:
        if not is_git_alias_setup(key, git_aliases[key], local=local):
            return False
    return True


git_aliases = {'pr': 'repos pr',
               'repo': 'repos'}



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# Generated file, DO NOT EDIT
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------------------------

from msrest import Serializer, Deserializer
from ...client import Client
from . import models


class ProfileClient(Client):
    """Profile
    :param str base_url: Service URL
    :param Authentication creds: Authenticated credentials.
    """

    def __init__(self, base_url=None, creds=None):
        super(ProfileClient, self).__init__(base_url, creds)
        client_models = {k: v for k, v in models.__dict__.items() if isinstance(v, type)}
        self._serialize = Serializer(client_models)
        self._deserialize = Deserializer(client_models)

    resource_area_identifier = '8ccfef3d-2b87-4e99-8ccb-66e343d2daa8'

    def get_profile(self, id, details=None, with_attributes=None, partition=None, core_attributes=None, force_refresh=None):
        """GetProfile.
        Get my profile.
        :param str id:
        :param bool details:
        :param bool with_attributes:
        :param str partition:
        :param str core_attributes:
        :param bool force_refresh:
        :rtype: :class:`<Profile> <azure.devops.v5_0.profile.models.Profile>`
        """
        route_values = {}
        if id is not None:
            route_values['id'] = self._serialize.url('id', id, 'str')
        query_parameters = {}
        if details is not None:
            query_parameters['details'] = self._serialize.query('details', details, 'bool')
        if with_attributes is not None:
            query_parameters['withAttributes'] = self._serialize.query('with_attributes', with_attributes, 'bool')
        if partition is not None:
            query_parameters['partition'] = self._serialize.query('partition', partition, 'str')
        if core_attributes is not None:
            query_parameters['coreAttributes'] = self._serialize.query('core_attributes', core_attributes, 'str')
        if force_refresh is not None:
            query_parameters['forceRefresh'] = self._serialize.query('force_refresh', force_refresh, 'bool')
        response = self._send(http_method='GET',
                              location_id='f83735dc-483f-4238-a291-d45f6080a9af',
                              version='5.0',
                              route_values=route_values,
                              query_parameters=query_parameters)
        return self._deserialize('Profile', response)




# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from knack.util import CLIError
try:
    # Attempt to load mock (works on Python 3.3 and above)
    from unittest.mock import patch
except ImportError:
    # Attempt to load mock (works on Python version below 3.3)
    from mock import patch

from azext_devops.dev.repos.policy import *
from azext_devops.devops_sdk.v5_0.policy.models import PolicyConfiguration, PolicyTypeRef
from azext_devops.dev.common.services import clear_connection_cache
from azext_devops.tests.utils.authentication import AuthenticatedTests
from azext_devops.devops_sdk.v5_0.policy.policy_client import PolicyClient


class TestUuidMethods(AuthenticatedTests):

    _TEST_DEVOPS_ORGANIZATION = 'https://someorg.visualstudio.com'
    _TEST_DEVOPS_PROJECT = 'sample project'
    _TEST_PAT_TOKEN = 'lwghjbj67fghokrgxsytghg75nk2ssguljk7a78qpcg2ttygviyt'
    _TEST_REPOSITORY_ID = 'b4da517c-0398-42dc-b2a8-0d3f240757f9'

    def setUp(self):
        self.authentication_setup()
        self.get_client = patch('azext_devops.devops_sdk.connection.Connection.get_client')
        self.get_policies_patcher = patch('azext_devops.devops_sdk.v5_0.policy.policy_client.PolicyClient.get_policy_configurations')
        self.get_policy_patcher = patch('azext_devops.devops_sdk.v5_0.policy.policy_client.PolicyClient.get_policy_configuration')
        self.delete_policy_patcher = patch('azext_devops.devops_sdk.v5_0.policy.policy_client.PolicyClient.delete_policy_configuration')
        self.create_policy_patcher = patch('azext_devops.devops_sdk.v5_0.policy.policy_client.PolicyClient.create_policy_configuration')
        self.update_policy_patcher = patch('azext_devops.devops_sdk.v5_0.policy.policy_client.PolicyClient.update_policy_configuration')

        self.mock_get_client = self.get_client.start()
        self.mock_get_policies = self.get_policies_patcher.start()
        self.mock_get_policy = self.get_policy_patcher.start()
        self.mock_delete_policy = self.delete_policy_patcher.start()
        self.mock_create_policy = self.create_policy_patcher.start()
        self.mock_update_policy = self.update_policy_patcher.start()

        self.mock_get_client.return_value = PolicyClient(base_url=self._TEST_DEVOPS_ORGANIZATION)

        clear_connection_cache()

    def tearDown(self):
        patch.stopall()

    def test_list_policy(self):
        list_policy(organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT)

        #assert
        self.mock_get_policies.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, scope=None)

    def test_list_policy_repo_scope(self):
        list_policy(organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT,
        repository_id='fake_repo_id')

        #assert
        self.mock_get_policies.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, scope='fake_repo_id')

    def test_list_policy_branch_scope(self):
        list_policy(organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT,
        repository_id='1d1dad71-f27c-4370-810d-838ec41efd41',
        branch='master')

        #assert
        self.mock_get_policies.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, scope='1d1dad71f27c4370810d838ec41efd41:refs/heads/master')

    def test_list_policy__only_branch_scope_error(self):
        try:
            list_policy(organization = self._TEST_DEVOPS_ORGANIZATION,
            project = self._TEST_DEVOPS_PROJECT,
            branch='master')
            self.fail('failure was expected')
        except CLIError as ex:
            #assert
            self.assertEqual(str(ex),
            '--repository-id is required with --branch')

    def test_get_policy(self):
        get_policy(policy_id = 121,
        organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT)

        #assert
        self.mock_get_policy.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, configuration_id=121)

    def test_delete_policy(self):
        delete_policy(policy_id = 121,
        organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT)

        #assert
        self.mock_delete_policy.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, configuration_id=121)

    def test_create_policy_config_file(self):
        configuration = {
            'is_blocking' : True,
            'is_enabled' : False
        }
        import tempfile
        temp_config_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        import json
        json.dump(configuration, temp_config_file)
        temp_config_file.close()

        create_policy_configuration_file(policy_configuration = temp_config_file.name,
        organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT)

        #assert
        self.mock_create_policy.assert_called_once()
        create_policy_object = self.mock_create_policy.call_args_list[0][1]
        self.assertEqual(self._TEST_DEVOPS_PROJECT, create_policy_object['project'], str(create_policy_object))
        self.assertEqual(create_policy_object['configuration']['is_enabled'], False)
        self.assertEqual(create_policy_object['configuration']['is_blocking'], True)

    def test_update_policy_configuration_file(self):
        configuration = {
            'is_blocking' : True,
            'is_enabled' : False
        }
        import tempfile
        temp_config_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        import json
        json.dump(configuration, temp_config_file)
        temp_config_file.close()

        update_policy_configuration_file(policy_id = 121,
        policy_configuration = temp_config_file.name,
        organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT)

        #assert
        self.mock_update_policy.assert_called_once()
        update_policy_object = self.mock_update_policy.call_args_list[0][1]
        self.assertEqual(update_policy_object['configuration_id'], 121)
        self.assertEqual(self._TEST_DEVOPS_PROJECT, update_policy_object['project'])
        self.assertEqual(update_policy_object['configuration']['is_enabled'], False)
        self.assertEqual(update_policy_object['configuration']['is_blocking'], True)

    def test_create_policy_approver_count(self):
        create_policy_approver_count(repository_id = self._TEST_REPOSITORY_ID,
        branch='master',
        blocking=False,
        enabled=True,
        minimum_approver_count='5',
        creator_vote_counts=False,
        allow_downvotes=False,
        reset_on_source_push=True,
        organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT)

        #assert
        self.mock_create_policy.assert_called_once()
        self.mock_create_policy.assert_called_once()
        create_policy_object = self.mock_create_policy.call_args_list[0][1]
        self.assertEqual(self._TEST_DEVOPS_PROJECT, create_policy_object['project'])
        settings = create_policy_object['configuration'].settings
        self.assertEqual(settings['minimumApproverCount'], '5')
        self.assertEqual(settings['creatorVoteCounts'], False)
        self.assertEqual(settings['allowDownvotes'], False)
        self.assertEqual(settings['resetOnSourcePush'], True)
        scope = create_policy_object['configuration'].settings['scope'][0]  # 0 because we set only only scope from CLI
        self.assertEqual(scope['repositoryId'], self._TEST_REPOSITORY_ID)
        self.assertEqual(scope['refName'], 'refs/heads/master')
        self.assertEqual(scope['matchKind'], 'exact')

    def test_update_policy_approver_count(self):
        current_policy = PolicyConfiguration(is_blocking=False, is_enabled=False)
        policy_type = PolicyTypeRef()
        policy_type.id = 'fa4e907d-c16b-4a4c-9dfa-4906e5d171dd'
        current_policy.type = policy_type
        current_policy.settings = {
            'minimumApproverCount' : 2,
            'creatorVoteCounts' : False,
            'allowDownvotes' : False,
            'resetOnSourcePush' : False,
            'scope':[
                {
                    'refName': 'ref\heads\master',
                    'repositoryId':self._TEST_REPOSITORY_ID,
                    'matchKind': 'exact'
                        }
                        ]
            }

        self.mock_get_policy.return_value = current_policy

        update_policy_approver_count(policy_id=121,
        allow_downvotes=True,
        blocking=True,
        reset_on_source_push=False,
        organization = self._TEST_DEVOPS_ORGANIZATION,
        project = self._TEST_DEVOPS_PROJECT)

        self.mock_get_policy.assert_called_once()

        self.mock_update_policy.assert_called_once()
        update_policy_object = self.mock_update_policy.call_args_list[0][1]
        self.assertEqual(update_policy_object['configuration_id'], 121)
        self.assertEqual(update_policy_object['configuration'].is_enabled, False)
        self.assertEqual(update_policy_object['configuration'].is_blocking, True)
        settings = update_policy_object['configuration'].settings
        self.assertEqual(settings['minimumApproverCount'], 2)
        self.assertEqual(settings['creatorVoteCounts'], False)
        self.assertEqual(settings['allowDownvotes'], True)
        self.assertEqual(settings['resetOnSourcePush'], False)





if __name__ == '__main__':
    unittest.main()


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from knack.log import get_logger
from knack.util import CLIError

logger = get_logger(__name__)

FILE_ENCODING_TYPES = ['ascii', 'utf-16be', 'utf-16le', 'utf-8']


def read_file_content(file_path, encoding):
    if not file_path or not encoding:
        raise CLIError("File path {} or encoding {} is missing.".format(file_path, encoding))

    if encoding not in FILE_ENCODING_TYPES:
        raise CLIError("File encoding {encoding} is not supported.".format(encoding=encoding))

    try:
        import sys
        if sys.version_info[0] < 3:
            return _read_file_content_ver2(file_path, encoding)

        return _read_file_content_ver3(file_path, encoding)
    except UnicodeDecodeError as ex:
        logger.debug(msg=ex)
        raise CLIError("Unable to decode file '{}' with '{}' encoding.".format(
            file_path, encoding))


def open_file(filepath):
    """
    Opens a file in the default editor for the file type and exits.
    """
    import subprocess
    import platform
    import os
    if platform.system() == 'Darwin':       # macOS
        subprocess.call(('open', filepath))
    elif platform.system() == 'Windows':    # Windows
        os.system(filepath)
    else:                                   # linux variants
        subprocess.call(('xdg-open', filepath))


def delete_dir(path):
    import shutil
    shutil.rmtree(path)


def datetime_now_as_string():
    from datetime import datetime
    now = datetime.utcnow().isoformat()
    return ''.join(e for e in now if e.isalnum())


def open_url(url):
    """Opens the url in new window in the default browser.
    """
    from webbrowser import open_new
    open_new(url=url)


# Decorators
def singleton(myclass):
    instance = [None]

    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = myclass(*args, **kwargs)
        return instance[0]
    return wrapper


def _read_file_content_ver3(file_path, encoding):
    logger.debug('inside read_file_content_ver3')
    with open(file_path, 'r', encoding=encoding) as f:
        return f.read()


def _read_file_content_ver2(file_path, encoding):
    logger.debug('inside read_file_content_ver2')
    with open(file_path) as f:
        return f.read().decode(encoding)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest

try:
    # Attempt to load mock (works on Python 3.3 and above)
    from unittest.mock import patch
except ImportError:
    # Attempt to load mock (works on Python version below 3.3)
    from mock import patch

from azext_devops.dev.pipelines.runs_artifacts import run_artifact_download, run_artifact_upload, run_artifact_list
from azext_devops.dev.common.services import clear_connection_cache
from azext_devops.tests.utils.authentication import AuthenticatedTests
from azext_devops.devops_sdk.v5_0.build.build_client import BuildClient
from azext_devops.dev.common.const import ARTIFACTTOOL_PAT_ENVKEY


class TestPipelinesRunsArtifactsMethods(AuthenticatedTests):

    _TEST_DEVOPS_ORGANIZATION = 'https://someorganization.visualstudio.com'
    _TEST_DEVOPS_PROJECT = 'MyProject'
    _TEST_PAT_TOKEN = 'Some_PAT_Token'
    _TEST_ARTIFACT_NAME = 'artifactname'
    _TEST_ARTIFACT_PATH = 'D:/Testpath'

    def setUp(self):
        self.authentication_setup()
        self.authenticate()
        self.get_client_patcher = patch('azext_devops.devops_sdk.connection.Connection.get_client')
        self.get_artifacts_patcher = patch('azext_devops.devops_sdk.v5_0.build.build_client.BuildClient.get_artifacts')
        self.run_artifacttool_patcher = patch('azext_devops.dev.common.artifacttool.ArtifactToolInvoker.run_artifacttool')

        #start the patchers
        self.mock_get_artifacts = self.get_artifacts_patcher.start()
        self.mock_get_client = self.get_client_patcher.start()
        self.mock_run_artifacttool = self.run_artifacttool_patcher.start()

        # Set return values which will be same across tests
        self.mock_get_client.return_value = BuildClient(base_url=self._TEST_DEVOPS_ORGANIZATION)

        #clear connection cache before running each test
        clear_connection_cache()

    def tearDown(self):
        patch.stopall()

    def test_runs_artifacts_list(self):
        # set return values
        response = run_artifact_list(run_id=12345, organization=self._TEST_DEVOPS_ORGANIZATION, 
            project=self._TEST_DEVOPS_PROJECT, detect=None)
        #assert
        self.mock_get_artifacts.assert_called_once_with(build_id=12345, project=self._TEST_DEVOPS_PROJECT)


    def test_runs_artifacts_download(self):
        # set return values
        response = run_artifact_download(
            run_id=12345, artifact_name=self._TEST_ARTIFACT_NAME, path=self._TEST_ARTIFACT_PATH,
            organization=self._TEST_DEVOPS_ORGANIZATION, project=self._TEST_DEVOPS_PROJECT, detect=None)
        #assert
        self.mock_run_artifacttool.assert_called_with(self._TEST_DEVOPS_ORGANIZATION,
            [
                'pipelineartifact',
                'download',
                '--service', self._TEST_DEVOPS_ORGANIZATION,
                '--patvar', ARTIFACTTOOL_PAT_ENVKEY,
                '--project', self._TEST_DEVOPS_PROJECT,
                '--pipeline-id', 12345,
                '--artifact-name', self._TEST_ARTIFACT_NAME,
                '--path', self._TEST_ARTIFACT_PATH,
            ], 
            'Downloading')

    def test_runs_artifacts_upload(self):
        # set return values
        response = run_artifact_upload(
            run_id=12345, artifact_name=self._TEST_ARTIFACT_NAME, path=self._TEST_ARTIFACT_PATH,
            organization=self._TEST_DEVOPS_ORGANIZATION, project=self._TEST_DEVOPS_PROJECT, detect=None)
        #assert
        self.mock_run_artifacttool.assert_called_with(self._TEST_DEVOPS_ORGANIZATION,
            [
                'pipelineartifact',
                'publish',
                '--service', self._TEST_DEVOPS_ORGANIZATION,
                '--patvar', ARTIFACTTOOL_PAT_ENVKEY,
                '--project', self._TEST_DEVOPS_PROJECT,
                '--pipeline-id', 12345,
                '--artifact-name', self._TEST_ARTIFACT_NAME,
                '--path', self._TEST_ARTIFACT_PATH,
            ], 
            'Uploading')

if __name__ == '__main__':
    unittest.main()


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azext_devops.dev.common.arguments import convert_date_string_to_iso8601
from azext_devops.dev.team.arguments import load_global_args


# pylint: disable=too-many-statements
def load_migration_arguments(self, _):
    with self.argument_context('devops migrations') as context:
        load_global_args(context)
        context.argument('repository_id', options_list='--repository-id',
                         help='ID of the Azure Repos repository (GUID).')

    with self.argument_context('devops migrations list') as context:
        context.argument('include_inactive', options_list='--include-inactive', action='store_true',
                         help='Include inactive (completed, abandoned, failed) migrations in the results.')
        context.argument('project', options_list='--project',
                         help='Optional project name or ID to filter migrations.')

    with self.argument_context('devops migrations create') as context:
        context.argument('target_repository', options_list='--target-repository',
                         help='Target repository URL (must start with http:// or https://).')
        context.argument('target_owner_user_id', options_list='--target-owner-user-id',
                         help='Target repository owner user ID.')
        context.argument('validate_only', options_list='--validate-only', action='store_true',
                         help='Create in validate-only mode (pre-migration checks only).')
        context.argument('cutover_date', options_list='--cutover-date',
                         type=convert_date_string_to_iso8601,
                         help='Scheduled cutover date/time (ISO 8601).')
        context.argument('agent_pool', options_list='--agent-pool',
                         help='Agent pool name to use for migration work.')
        context.argument('skip_validation', options_list='--skip-validation',
                         help='Validation policies to skip. Accepts either a comma-separated list of '
                              'policy names (for example, AgentPoolExists,MaxRepoSize) or a non-negative '
                              'integer bitmask.')

    with self.argument_context('devops migrations cutover set') as context:
        context.argument('cutover_date', options_list='--date',
                         type=convert_date_string_to_iso8601,
                         help='The date and time for cutover (ISO 8601).')

    with self.argument_context('devops migrations resume') as context:
        context.argument('validate_only', options_list='--validate-only', action='store_true',
                         help='Resume in validate-only mode.')
        context.argument('migration', options_list='--migration', action='store_true',
                         help='Promote a succeeded validate-only migration to a full migration '
                              '(sets validateOnly=false and statusRequested=active).')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import time

from .services import get_operations_client


def wait_for_long_running_operation(organization, operation_id, interval_seconds=5):
    operation_client = get_operations_client(organization)
    operation = operation_client.get_operation(operation_id)
    while not has_operation_completed(operation):
        time.sleep(interval_seconds)
        operation = operation_client.get_operation(operation_id)
    return operation


def has_operation_completed(operation):
    status = operation.status.lower()
    return status in ('succeeded', 'failed', 'cancelled')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# Generated file, DO NOT EDIT
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------------------------

from .models import *
from .tfvc_client import TfvcClient

__all__ = [
    'AssociatedWorkItem',
    'Change',
    'CheckinNote',
    'FileContentMetadata',
    'GitRepository',
    'GitRepositoryRef',
    'GraphSubjectBase',
    'IdentityRef',
    'ItemContent',
    'ItemModel',
    'ReferenceLinks',
    'TeamProjectCollectionReference',
    'TeamProjectReference',
    'TfvcBranch',
    'TfvcBranchMapping',
    'TfvcBranchRef',
    'TfvcChange',
    'TfvcChangeset',
    'TfvcChangesetRef',
    'TfvcChangesetSearchCriteria',
    'TfvcChangesetsRequestData',
    'TfvcItem',
    'TfvcItemDescriptor',
    'TfvcItemRequestData',
    'TfvcLabel',
    'TfvcLabelRef',
    'TfvcLabelRequestData',
    'TfvcMappingFilter',
    'TfvcMergeSource',
    'TfvcPolicyFailureInfo',
    'TfvcPolicyOverrideInfo',
    'TfvcShallowBranchRef',
    'TfvcShelveset',
    'TfvcShelvesetRef',
    'TfvcShelvesetRequestData',
    'TfvcStatistics',
    'TfvcVersionDescriptor',
    'VersionControlProjectInfo',
    'VstsInfo',
    'TfvcClient'
]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# Generated file, DO NOT EDIT
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------------------------

from .models import *
from .service_hooks_client import ServiceHooksClient

__all__ = [
    'Consumer',
    'ConsumerAction',
    'Event',
    'EventTypeDescriptor',
    'ExternalConfigurationDescriptor',
    'FormattedEventMessage',
    'GraphSubjectBase',
    'IdentityRef',
    'InputDescriptor',
    'InputFilter',
    'InputFilterCondition',
    'InputValidation',
    'InputValue',
    'InputValues',
    'InputValuesError',
    'InputValuesQuery',
    'Notification',
    'NotificationDetails',
    'NotificationResultsSummaryDetail',
    'NotificationsQuery',
    'NotificationSummary',
    'Publisher',
    'PublisherEvent',
    'PublishersQuery',
    'ReferenceLinks',
    'ResourceContainer',
    'SessionToken',
    'Subscription',
    'SubscriptionDiagnostics',
    'SubscriptionsQuery',
    'SubscriptionTracing',
    'UpdateSubscripitonDiagnosticsParameters',
    'UpdateSubscripitonTracingParameters',
    'VersionedResource',
    'ServiceHooksClient'
]



# coding=utf-8
# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import print_function

import os

from knack.config import get_config_parser
from knack.log import get_logger
from knack.util import CLIError

from azext_devops.dev.common.config import (set_global_config_value, AZ_DEVOPS_GLOBAL_CONFIG_PATH)
from azext_devops.dev.common.const import (CLI_ENV_VARIABLE_PREFIX, DEFAULTS_SECTION, DEVOPS_ORGANIZATION_DEFAULT,
                                           DEVOPS_TEAM_PROJECT_DEFAULT)
from azext_devops.dev.common.uri import is_valid_url

logger = get_logger(__name__)


CONFIG_VALID_DEFAULT_KEYS_LIST = [DEVOPS_ORGANIZATION_DEFAULT, DEVOPS_TEAM_PROJECT_DEFAULT]


def configure(defaults=None, use_git_aliases=None, list_config=False):
    """Configure the Azure DevOps CLI or view your configuration.
    :param defaults: Space separated 'name=value' pairs for common arguments defaults,
        e.g. '--defaults project=my-project-name organization=my-org-url
        arg=value'. Use '' to clear the defaults, e.g. --defaults project=''.
    :type defaults: str
    :param use_git_aliases: Set to 'true' to configure Git aliases global git config file
        (to enable commands like "git pr list").
        Set to 'false' to remove any aliases set by the tool.
    :type use_git_aliases: str
    :param list_config: Lists the contents of the config file.
    :type list_config: bool
    """
    if defaults is None and use_git_aliases is None and list_config is False:
        raise CLIError('usage error: atleast one of the options must be specified.'
                       'For list of supported options see help using -h flag.')
    if defaults:
        for default in defaults:
            parts = default.split('=', 1)
            if len(parts) == 1:
                raise CLIError('usage error: --defaults STRING=STRING STRING=STRING ...')
            if parts[0] not in CONFIG_VALID_DEFAULT_KEYS_LIST:
                raise CLIError('usage error: invalid default value setup. Supported values are {}.'
                               .format(CONFIG_VALID_DEFAULT_KEYS_LIST))
            _validate_configuration(key=parts[0], value=parts[1])
            set_global_config_value(section=DEFAULTS_SECTION, option=parts[0], value=parts[1])
    if use_git_aliases is not None:
        from azext_devops.dev.repos.git_alias import setup_git_aliases, clear_git_aliases
        if use_git_aliases:
            setup_git_aliases()
        else:
            clear_git_aliases()
    if list_config:
        print_current_configuration()


def print_current_configuration(file_config=None):
    from azext_devops.dev.repos.git_alias import are_git_aliases_setup
    if file_config is None:
        file_config = get_config_parser()
        file_config.read([AZ_DEVOPS_GLOBAL_CONFIG_PATH])
    for section in file_config.sections():
        print()
        print('[{}]'.format(section))
        for option in file_config.options(section):
            print('{} = {}'.format(option, file_config.get(section, option)))
    # Print if git alias is setup or not
    is_git_alias_setup = MSG_NO
    if are_git_aliases_setup():
        is_git_alias_setup = MSG_YES
    print('{} = {}'.format(MSG_GIT_ALIAS_SETUP, is_git_alias_setup))
    env_vars = [ev for ev in os.environ if ev.startswith(CLI_ENV_VARIABLE_PREFIX)]
    if env_vars:
        print(MSG_HEADING_ENV_VARS)
        print('\n'.join(['{}'.format(ev) for ev in env_vars]))


def _validate_configuration(key, value):
    if key == DEVOPS_ORGANIZATION_DEFAULT:
        # value can be '' or a valid url
        if not value == '' and not is_valid_url(value):
            raise CLIError('Organization should be a valid Azure DevOps or Azure DevOps Server repository url. '
                           'See command help for details.')


MSG_INTRO = '\nWelcome to the Azure DevOps CLI! This command will guide you through setting some default values.\n'
MSG_CLOSING = '\nYou\'re all set! Here are some commands to try:\n' \
              ' $ az devops login\n' \
              ' $ az repos pr list\n{}' \
              ' $ az devops feedback\n'

MSG_CLOSING_GIT_COMMAND = ' $ git pr list\n'
MSG_GIT_ALIAS_SETUP = '\nUse git alias'

MSG_GLOBAL_SETTINGS_LOCATION = 'Your settings can be found at {}'

MSG_HEADING_CURRENT_CONFIG_INFO = 'Your current configuration is as follows:'
MSG_HEADING_ENV_VARS = '\nEnvironment variables:'

MSG_PROMPT_MANAGE_GLOBAL = '\nDo you wish to change your settings?'
MSG_PROMPT_GIT_ALIAS = '\nConfigure aliases for Git (to enable commands like "git pr list")?'
MSG_PROMPT_CLEAR_GIT_ALIAS = '\nGit aliases are configured, would you like to clear them?'

MSG_YES = 'Yes'
MSG_NO = 'No'



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from collections import OrderedDict
import dateutil

_VALUE_TRUNCATION_LENGTH = 80


def transform_builds_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_build_row(item))
    return table_output


def transform_build_table_output(result):
    table_output = [_transform_build_row(result)]
    return table_output


def _transform_build_row(row):
    from azext_devops.dev.common.git import REF_HEADS_PREFIX
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Number'] = row['buildNumber']
    table_row['Status'] = row['status']
    if row['result']:
        table_row['Result'] = row['result']
    else:
        table_row['Result'] = ' '
    table_row['Definition ID'] = row['definition']['id']
    table_row['Definition Name'] = row['definition']['name']

    if row['sourceBranch']:
        source_branch = row['sourceBranch']
        if source_branch[0:len(REF_HEADS_PREFIX)] == REF_HEADS_PREFIX:
            source_branch = source_branch[len(REF_HEADS_PREFIX):]
        table_row['Source Branch'] = source_branch
    else:
        table_row['Source Branch'] = ' '

    queued_time = dateutil.parser.parse(row['queueTime']).astimezone(dateutil.tz.tzlocal())
    table_row['Queued Time'] = str(queued_time.date()) + ' ' + str(queued_time.time())
    table_row['Reason'] = row['reason']
    return table_row


def transform_build_tags_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_build_tags_row(item))
    return table_output


def _transform_build_tags_row(row):
    table_row = OrderedDict()
    table_row['Tags'] = row
    return table_row


def transform_definitions_table_output(result):
    table_output = []
    include_draft_column = False
    for item in result:
        if item['quality'] == 'draft':
            include_draft_column = True
            break
    for item in result:
        table_output.append(_transform_definition_row(item, include_draft_column))
    return table_output


def transform_definition_table_output(result):
    table_output = [_transform_definition_row(result, result['quality'] == 'draft')]
    return table_output


def _transform_definition_row(row, include_draft_column=False):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    if include_draft_column:
        if row['quality'] == 'draft':
            table_row['Draft'] = True
        else:
            table_row['Draft'] = ' '
    if row['queueStatus']:
        table_row['Status'] = row['queueStatus']
    else:
        table_row['Status'] = ' '
    if row['queue']:
        table_row['Default Queue'] = row['queue']['name']
    else:
        table_row['Default Queue'] = ' '
    return table_row


def _get_task_key(row):
    return row['name'].lower()


def transform_releases_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_release_row(item))
    return table_output


def transform_release_table_output(result):
    table_output = [_transform_release_row(result)]
    return table_output


def _transform_release_row(row):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    table_row['Definition Name'] = row['releaseDefinition']['name']
    table_row['Created By'] = row['createdBy']['displayName']
    created_on = dateutil.parser.parse(row['createdOn']).astimezone(dateutil.tz.tzlocal())
    table_row['Created On'] = str(created_on.date()) + ' ' + str(created_on.time())
    table_row['Status'] = row['status']
    table_row['Description'] = row['description']
    return table_row


def transform_release_definitions_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_release_definition_row(item))
    return table_output


def transform_release_definition_table_output(result):
    table_output = [_transform_release_definition_row(result)]
    return table_output


def _transform_release_definition_row(row):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    table_row['CreatedBy'] = row['createdBy']['displayName']
    table_row['Created On'] = row['createdOn']
    return table_row


def transform_runs_artifact_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_runs_artifact_row(item))
    return table_output


def _transform_runs_artifact_row(row):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    table_row['Type'] = row['resource']['type']
    return table_row


def transform_pipelines_table_output(result):
    table_output = []
    include_draft_column = False
    for item in result:
        if item['quality'] == 'draft':
            include_draft_column = True
            break
    for item in result:
        table_output.append(_transform_pipeline_row(item, include_draft_column))
    return table_output


def transform_pipeline_table_output(result):
    table_output = [_transform_pipeline_row(result, result['quality'] == 'draft')]
    return table_output


def _transform_pipeline_row(row, include_draft_column=False):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    path = row['path']
    table_row['Path'] = path[:50] + '..' if len(path) > 50 else path
    name = row['name']
    table_row['Name'] = name[:50] + '..' if len(name) > 50 else name
    if include_draft_column:
        if row['quality'] == 'draft':
            table_row['Draft'] = True
        else:
            table_row['Draft'] = ' '
    if row['queueStatus']:
        table_row['Status'] = row['queueStatus']
    else:
        table_row['Status'] = ' '
    if row['queue']:
        table_row['Default Queue'] = row['queue']['name']
    else:
        table_row['Default Queue'] = ' '
    return table_row


def transform_pipeline_runs_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_pipeline_run_row(item))
    return table_output


def transform_pipeline_or_run_table_output(result):
    table_output = [_transform_pipeline_or_run_row(result)]
    return table_output


def _transform_pipeline_or_run_row(row):
    if row.get("buildNumber", None):  # Hack to detect the json object is definition or run
        return _transform_pipeline_run_row(row)
    return _transform_pipeline_row(row)


def transform_pipeline_run_table_output(result):
    table_output = [_transform_pipeline_run_row(result)]
    return table_output


def _transform_pipeline_run_row(row):
    from azext_devops.dev.common.git import REF_HEADS_PREFIX
    table_row = OrderedDict()
    table_row['Run ID'] = row['id']
    table_row['Number'] = row['buildNumber']
    table_row['Status'] = row['status']
    if row['result']:
        table_row['Result'] = row['result']
    else:
        table_row['Result'] = ' '
    table_row['Pipeline ID'] = row['definition']['id']
    table_row['Pipeline Name'] = row['definition']['name']

    if row['sourceBranch']:
        source_branch = row['sourceBranch']
        if source_branch[0:len(REF_HEADS_PREFIX)] == REF_HEADS_PREFIX:
            source_branch = source_branch[len(REF_HEADS_PREFIX):]
        table_row['Source Branch'] = source_branch
    else:
        table_row['Source Branch'] = ' '

    queued_time = dateutil.parser.parse(row['queueTime']).astimezone(dateutil.tz.tzlocal())
    table_row['Queued Time'] = str(queued_time.date()) + ' ' + str(queued_time.time())
    table_row['Reason'] = row['reason']
    return table_row


def transform_pipelines_pools_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_pipeline_pool_row(item))
    return table_output


def transform_pipelines_pool_table_output(result):
    table_output = [_transform_pipeline_pool_row(result)]
    return table_output


def _transform_pipeline_pool_row(row):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    table_row['Is Hosted'] = row['isHosted']
    table_row['Pool Type'] = row['poolType']
    return table_row


def transform_pipelines_agents_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_pipeline_agent_row(item))
    return table_output


def transform_pipelines_agent_table_output(result):
    table_output = [_transform_pipeline_agent_row(result)]
    return table_output


def _transform_pipeline_agent_row(row):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    table_row['Is Enabled'] = row['enabled']
    table_row['Status'] = row['status']
    table_row['Version'] = row['version']
    return table_row


def transform_pipelines_queues_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_pipeline_queue_row(item))
    return table_output


def transform_pipelines_queue_table_output(result):
    table_output = [_transform_pipeline_queue_row(result)]
    return table_output


def _transform_pipeline_queue_row(row):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    table_row['Pool IsHosted'] = row['pool']['isHosted']
    table_row['Pool Type'] = row['pool']['poolType']
    return table_row


def transform_pipelines_variable_groups_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_pipeline_variable_group_row(item))
    return table_output


def transform_pipelines_variable_group_table_output(result):
    table_output = [_transform_pipeline_variable_group_row(result)]
    return table_output


def _transform_pipeline_variable_group_row(row):
    table_row = OrderedDict()
    table_row['ID'] = row['id']
    table_row['Name'] = row['name']
    table_row['Type'] = row['type']
    table_row['Description'] = row['description']
    if row.get('authorized', None) is not None:
        table_row['Is Authorized'] = row['authorized']
    table_row['Number of Variables'] = len(row['variables'])
    return table_row


def transform_pipelines_variables_table_output(result):
    table_output = []
    for key, value in result.items():
        table_output.append(_transform_pipeline_variable_row(key, value))
    return table_output


def _transform_pipeline_variable_row(key, value):
    table_row = OrderedDict()
    table_row['Name'] = key
    table_row['Allow Override'] = 'True' if value['allowOverride'] else 'False'
    table_row['Is Secret'] = 'True' if value['isSecret'] else 'False'
    val = value['value'] if value['value'] is not None else ''
    if len(val) > _VALUE_TRUNCATION_LENGTH:
        val = val[0:_VALUE_TRUNCATION_LENGTH] + '...'
    table_row['Value'] = val
    return table_row


def transform_pipelines_var_group_variables_table_output(result):
    table_output = []
    for key, value in result.items():
        table_output.append(_transform_pipeline_var_group_variable_row(key, value))
    return table_output


def _transform_pipeline_var_group_variable_row(key, value):
    table_row = OrderedDict()
    table_row['Name'] = key
    table_row['Is Secret'] = 'True' if value['isSecret'] else 'False'
    val = value['value'] if value['value'] is not None else ''
    if len(val) > _VALUE_TRUNCATION_LENGTH:
        val = val[0:_VALUE_TRUNCATION_LENGTH] + '...'
    table_row['Value'] = val
    return table_row


def transform_pipelines_folders_table_output(result):
    table_output = []
    for item in result:
        table_output.append(_transform_pipeline_folder_row(item))
    return table_output


def transform_pipelines_folder_table_output(result):
    table_output = [_transform_pipeline_folder_row(result)]
    return table_output


def _transform_pipeline_folder_row(row):
    table_row = OrderedDict()
    table_row['Path'] = row['path']
    val = row['description'] if row['description'] is not None else ''
    if len(val) > _VALUE_TRUNCATION_LENGTH:
        val = val[0:_VALUE_TRUNCATION_LENGTH] + '...'
    table_row['Description'] = val
    return table_row



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import print_function

import os
from knack.log import get_logger
from knack.prompting import prompt_pass
from knack.util import CLIError
from azext_devops.devops_sdk.v5_0.service_endpoint.models import ServiceEndpoint, EndpointAuthorization
from azext_devops.dev.common.services import get_service_endpoint_client, resolve_instance_and_project
from azext_devops.dev.common.const import CLI_ENV_VARIABLE_PREFIX, AZ_DEVOPS_GITHUB_PAT_ENVKEY
from azext_devops.dev.common.prompting import verify_is_a_tty_or_raise_error

from .const import (SERVICE_ENDPOINT_AUTHORIZATION_PERSONAL_ACCESS_TOKEN,
                    SERVICE_ENDPOINT_TYPE_GITHUB,
                    SERVICE_ENDPOINT_AUTHORIZATION_SERVICE_PRINCIPAL,
                    SERVICE_ENDPOINT_TYPE_AZURE_RM)

logger = get_logger(__name__)


# pylint: disable=too-few-public-methods, too-many-instance-attributes
class ServiceEndpointAuthorized():
    _attribute_map = {
        'service_endpoint_parameters': {'key': 'service_endpoint_parameters', 'type': 'ServiceEndpoint'},
        'authorized': {'key': 'authorized', 'type': 'bool'}
    }

    def __init__(self, service_endpoint_parameters, authorized):
        self.authorized = authorized if authorized is not None else False
        self.administratorsGroup = service_endpoint_parameters.administrators_group
        self.authorization = service_endpoint_parameters.authorization
        self.createdBy = service_endpoint_parameters.created_by
        self.data = service_endpoint_parameters.data
        self.description = service_endpoint_parameters.description
        self.groupScopeId = service_endpoint_parameters.group_scope_id
        self.id = service_endpoint_parameters.id
        self.isReady = service_endpoint_parameters.is_ready
        self.isShared = service_endpoint_parameters.is_shared
        self.name = service_endpoint_parameters.name
        self.operationStatus = service_endpoint_parameters.operation_status
        self.owner = service_endpoint_parameters.owner
        self.readersGroup = service_endpoint_parameters.readers_group
        self.type = service_endpoint_parameters.type
        self.url = service_endpoint_parameters.url


def list_service_endpoints(organization=None, project=None, detect=None):
    """List service endpoints in a project.
    :rtype: list of :class:`VssJsonCollectionWrapper <service_endpoint.v4_1.models.ServiceEndpoint>`
    """
    organization, project = resolve_instance_and_project(detect=detect,
                                                         organization=organization,
                                                         project=project)
    client = get_service_endpoint_client(organization)
    return client.get_service_endpoints(project)


def show_service_endpoint(id, organization=None, project=None, detect=None):  # pylint: disable=redefined-builtin
    """Get the details of a service endpoint.
    :param id: ID of the service endpoint.
    :type id: str
    :rtype: :class:`ServiceEndpoint <service_endpoint.v4_1.models.ServiceEndpoint>`
    """
    organization, project = resolve_instance_and_project(detect=detect,
                                                         organization=organization,
                                                         project=project)
    client = get_service_endpoint_client(organization)
    return client.get_service_endpoint_details(project, id)


def delete_service_endpoint(id, deep=False, organization=None, project=None, detect=None):  # pylint: disable=redefined-builtin
    """Deletes service endpoint
    :param id: Id of the service endpoint to delete.
    :type id: str
    :param deep: Specific to AzureRM endpoint created in Automatic flow. When it is specified,
    this will also delete corresponding AAD application in Azure.
    :type deep: bool
    """
    organization, project = resolve_instance_and_project(detect=detect,
                                                         organization=organization,
                                                         project=project)
    client = get_service_endpoint_client(organization)
    return client.delete_service_endpoint(project, id, deep)


def create_azurerm_service_endpoint(name, azure_rm_tenant_id, azure_rm_service_principal_id,
                                    azure_rm_subscription_id, azure_rm_subscription_name,
                                    azure_rm_service_principal_certificate_path=None,
                                    organization=None, project=None, detect=None):
    """ Create an Azure RM type service endpoint.
    :param name: Name of service endpoint to create
    :type name: str
    :param azure_rm_tenant_id: tenant id for creating azure rm service endpoint
    :type azure_rm_tenant_id: str
    :param azure_rm_service_principal_id: service principal id for creating azure rm service endpoint
    :type azure_rm_service_principal_id: str
    :param azure_rm_subscription_id: subscription id for azure rm service endpoint
    :type azure_rm_subscription_id: str
    :param azure_rm_service_principal_certificate_path: Path to (.pem) which is certificate.
     Create using command "openssl pkcs12 -in file.pfx -out file.pem -nodes -password pass:password_here".
     More details : https://aka.ms/azure-devops-cli-azurerm-service-endpoint
    :type azure_rm_service_principal_certificate_path: str
    :param azure_rm_subscription_name: name of azure subscription for azure rm service endpoint
    :type azure_rm_subscription_name: str
    :rtype: :class:`ServiceEndpoint <service_endpoint.v4_1.models.ServiceEndpoint>`
    """
    organization, project = resolve_instance_and_project(detect=detect,
                                                         organization=organization,
                                                         project=project)
    client = get_service_endpoint_client(organization)

    service_endpoint_authorization = EndpointAuthorization(
        parameters={'tenantid': azure_rm_tenant_id,
                    'serviceprincipalid': azure_rm_service_principal_id},
        scheme=SERVICE_ENDPOINT_AUTHORIZATION_SERVICE_PRINCIPAL)

    if azure_rm_service_principal_certificate_path is None:
        AZURE_RM_SP_KEY_END_VARIABLE_NAME = CLI_ENV_VARIABLE_PREFIX + 'AZURE_RM_SERVICE_PRINCIPAL_KEY'
        if AZURE_RM_SP_KEY_END_VARIABLE_NAME not in os.environ:
            error_message = 'Please specify azure service principal key in ' + AZURE_RM_SP_KEY_END_VARIABLE_NAME +\
                            ' environment variable in non-interactive mode or use ' +\
                            '--azure-rm-service-principal-certificate-path.'
            verify_is_a_tty_or_raise_error(error_message)
            azure_rm_service_principal_key = prompt_pass('Azure RM service principal key:', confirm=True)
        else:
            logger.debug('Picking Azure RM principal key from environment variable')
            azure_rm_service_principal_key = os.environ[AZURE_RM_SP_KEY_END_VARIABLE_NAME]

        service_endpoint_authorization.parameters['authenticationType'] = 'spnKey'
        service_endpoint_authorization.parameters['serviceprincipalkey'] = azure_rm_service_principal_key
    else:
        with open(azure_rm_service_principal_certificate_path, "r") as f:
            service_endpoint_authorization.parameters['authenticationType'] = 'spnCertificate'
            service_endpoint_authorization.parameters['servicePrincipalCertificate'] = f.read()

    service_endpoint_data = {
        'subscriptionId': azure_rm_subscription_id,
        'subscriptionName': azure_rm_subscription_name,
        'environment': 'AzureCloud',
        'creationMode': 'Manual'
    }
    service_endpoint_to_create = ServiceEndpoint(
        authorization=service_endpoint_authorization, data=service_endpoint_data,
        name=name, type=SERVICE_ENDPOINT_TYPE_AZURE_RM, url='https://management.azure.com/')
    return client.create_service_endpoint(service_endpoint_to_create, project)


def create_github_service_endpoint(name, github_url,
                                   organization=None, project=None, detect=None):
    """ Create a GitHub service endpoint.
    :param name: Name of service endpoint to create
    :type name: str
    :param github_url: Url for github for creating service endpoint
    :type github_url: str
    :rtype: :class:`ServiceEndpoint <service_endpoint.v4_1.models.ServiceEndpoint>`
    """
    organization, project = resolve_instance_and_project(detect=detect,
                                                         organization=organization,
                                                         project=project)
    client = get_service_endpoint_client(organization)
    if AZ_DEVOPS_GITHUB_PAT_ENVKEY not in os.environ:
        error_message = 'Please pass GitHub access token in ' + AZ_DEVOPS_GITHUB_PAT_ENVKEY +\
                        ' environment variable in non-interactive mode.'
        verify_is_a_tty_or_raise_error(error_message)
        github_access_token = prompt_pass('GitHub access token:', confirm=True)
    else:
        logger.debug('Picking GitHub PAT from environment variable')
        github_access_token = os.environ[AZ_DEVOPS_GITHUB_PAT_ENVKEY]

    service_endpoint_authorization = EndpointAuthorization(
        parameters={'accessToken': github_access_token},
        scheme=SERVICE_ENDPOINT_AUTHORIZATION_PERSONAL_ACCESS_TOKEN)
    service_endpoint_to_create = ServiceEndpoint(
        authorization=service_endpoint_authorization,
        name=name, type=SERVICE_ENDPOINT_TYPE_GITHUB, url=github_url)
    return client.create_service_endpoint(service_endpoint_to_create, project)


def create_service_endpoint(service_endpoint_configuration,
                            encoding='utf-8', organization=None,
                            project=None, detect=None):
    """Create a service endpoint using configuration file.
    :param name: Name of service endpoint to create
    :type name: str
    :param service_endpoint_configuration: Configuration file with service endpoint request.
    :type service_endpoint_configuration: str
    :rtype: :class:`ServiceEndpoint <service_endpoint.v4_1.models.ServiceEndpoint>`
    """
    organization, project = resolve_instance_and_project(detect=detect,
                                                         organization=organization,
                                                         project=project)
    client = get_service_endpoint_client(organization)
    from azext_devops.dev.common.utils import read_file_content
    in_file_content = read_file_content(file_path=service_endpoint_configuration, encoding=encoding)
    import json
    service_endpoint_to_create = json.loads(in_file_content)
    return client.create_service_endpoint(service_endpoint_to_create, project)


def update_service_endpoint(id, enable_for_all=None, organization=None,  # pylint: disable=redefined-builtin
                            project=None, detect=None):
    """Update a service endpoint
    :param id: ID of the service endpoint.
    :type id: str
    """
    if enable_for_all is None:
        raise CLIError('Atleast one property to be updated must be specified.')
    organization, project = resolve_instance_and_project(detect=detect,
                                                         organization=organization,
                                                         project=project)
    client = get_service_endpoint_client(organization)
    se = client.get_service_endpoint_details(project, id)

    # set authorization if get service endpoint succeeded
    from azext_devops.dev.pipelines.pipeline_utils import set_authorize_resource, get_authorize_resource
    set_authorize_resource(
        authorized=enable_for_all, res_id=se.id, name=se.name, res_type='endpoint',
        organization=organization, project=project)

    authorized = get_authorize_resource(res_id=se.id, res_type='endpoint', organization=organization, project=project)
    return ServiceEndpointAuthorized(service_endpoint_parameters=se, authorized=authorized)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest

try:
    # Attempt to load mock (works on Python 3.3 and above)
    from unittest.mock import patch
except ImportError:
    # Attempt to load mock (works on Python version below 3.3)
    from mock import patch

from azext_devops.dev.common.services import clear_connection_cache
from azext_devops.tests.utils.authentication import AuthenticatedTests
from azext_devops.devops_sdk.v5_1.task_agent.task_agent_client import TaskAgentClient
from azext_devops.dev.pipelines.agent_pool_queue import list_pools, show_pool, list_agents, show_agent, list_queues, show_queue

class TestPipelinesAgentPoolQueueMethods(AuthenticatedTests):

    _TEST_DEVOPS_ORGANIZATION = 'https://someorganization.visualstudio.com'
    _TEST_DEVOPS_PROJECT = 'MyProject'

    def setUp(self):
        self.authentication_setup()
        self.authenticate()

        self.get_client_patcher = patch('azext_devops.devops_sdk.connection.Connection.get_client')
        #start the patchers
        self.mock_get_client = self.get_client_patcher.start()
        # Set return values which will be same across tests
        self.mock_get_client.return_value = TaskAgentClient(base_url=self._TEST_DEVOPS_ORGANIZATION)
        #clear connection cache before running each test
        clear_connection_cache()

    def tearDown(self):
        patch.stopall()

    def test_pool_list_with_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_pools') as mock_get_pools:
            # pool listing
            pools = list_pools(pool_name=None, pool_type=None, action=None, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_pools.assert_called_once_with(pool_name=None, properties=None, pool_type=None, action_filter=None)

    def test_pool_list_without_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_pools') as mock_get_pools:
            # pool listing
            pools = list_pools(pool_name='PoolName', pool_type='PoolType', action='ActionUse', organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_pools.assert_called_once_with(pool_name='PoolName', properties=None, pool_type='PoolType', action_filter='ActionUse')

    def test_pool_show_with_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_pool') as mock_get_pool:
            # pool show
            pool = show_pool(pool_id='123', action=None, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_pool.assert_called_once_with(pool_id='123', properties=None, action_filter=None)

    def test_pool_show_without_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_pool') as mock_get_pool:
            # pool show
            pool = show_pool(pool_id='123', action='ActionUse', organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_pool.assert_called_once_with(pool_id='123', properties=None, action_filter='ActionUse')

    def test_agent_list_with_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agents') as mock_get_agents:
            # agent listing
            agents = list_agents(pool_id='123', agent_name=None, include_capabilities=None, include_assigned_request=None,
                include_last_completed_request=None, demands=None, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_agents.assert_called_once_with(pool_id='123', agent_name=None, include_capabilities=None,
        include_last_completed_request=None, include_assigned_request=None, property_filters=None, demands=None)

    def test_agent_list_without_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agents') as mock_get_agents:
            # agent listing
            agents = list_agents(pool_id='123', agent_name='myagent', include_capabilities=True, include_assigned_request=True,
                include_last_completed_request=True, demands='demand1,demand2', organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_agents.assert_called_once_with(pool_id='123', agent_name='myagent', include_capabilities=True,
        include_last_completed_request=True, include_assigned_request=True, property_filters=None, demands=['demand1', 'demand2'])

    def test_agent_show_with_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent') as mock_get_agent:
            # agent Show
            agents = show_agent(pool_id='123', agent_id='1234', include_capabilities=None, include_assigned_request=None,
               include_last_completed_request=None, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_agent.assert_called_once_with(
                pool_id='123', agent_id='1234', include_capabilities=None,
                include_assigned_request=None, include_last_completed_request=None, property_filters=None)

    def test_agent_show_without_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent') as mock_get_agent:
            # agent Show
            agents = show_agent(pool_id='123', agent_id='1234', include_capabilities=True, include_assigned_request=True,
               include_last_completed_request=True, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_agent.assert_called_once_with(
                pool_id='123', agent_id='1234', include_capabilities=True,
                include_assigned_request=True, include_last_completed_request=True, property_filters=None)

    def test_queue_list_with_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_queues') as mock_get_queues:
            # queue listing
            queues = list_queues(queue_name=None, action=None, project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_queues.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, queue_name=None, action_filter=None)

    def test_queue_list_without_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_queues') as mock_get_queues:
            # queue listing
            queues = list_queues(queue_name='Ubuntu', action='ActionUse', project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_queues.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, queue_name='Ubuntu', action_filter='ActionUse')

    def test_queue_show_with_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_queue') as mock_get_queue:
            # queue listing
            queues = show_queue(queue_id='123', action=None, project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_queue.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, queue_id='123', action_filter=None)

    def test_queue_show_without_defaults(self):
        with patch('azext_devops.devops_sdk.v5_1.task_agent.task_agent_client.TaskAgentClient.get_agent_queue') as mock_get_queue:
            # queue listing
            queues = show_queue(queue_id='123', action='ActionUse', project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_get_queue.assert_called_once_with(project=self._TEST_DEVOPS_PROJECT, queue_id='123', action_filter='ActionUse')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# Generated file, DO NOT EDIT
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------------------------

from msrest.serialization import Model


class AddProcessWorkItemTypeFieldRequest(Model):
    """AddProcessWorkItemTypeFieldRequest.

    :param allow_groups: Allow setting field value to a group identity. Only applies to identity fields.
    :type allow_groups: bool
    :param default_value: The default value of the field.
    :type default_value: object
    :param read_only: If true the field cannot be edited.
    :type read_only: bool
    :param reference_name: Reference name of the field.
    :type reference_name: str
    :param required: If true the field cannot be empty.
    :type required: bool
    """

    _attribute_map = {
        'allow_groups': {'key': 'allowGroups', 'type': 'bool'},
        'default_value': {'key': 'defaultValue', 'type': 'object'},
        'read_only': {'key': 'readOnly', 'type': 'bool'},
        'reference_name': {'key': 'referenceName', 'type': 'str'},
        'required': {'key': 'required', 'type': 'bool'}
    }

    def __init__(self, allow_groups=None, default_value=None, read_only=None, reference_name=None, required=None):
        super(AddProcessWorkItemTypeFieldRequest, self).__init__()
        self.allow_groups = allow_groups
        self.default_value = default_value
        self.read_only = read_only
        self.reference_name = reference_name
        self.required = required


class Control(Model):
    """Control.

    :param contribution: Contribution for the control.
    :type contribution: :class:`WitContribution <azure.devops.v5_0.work_item_tracking.models.WitContribution>`
    :param control_type: Type of the control.
    :type control_type: str
    :param height: Height of the control, for html controls.
    :type height: int
    :param id: The id for the layout node.
    :type id: str
    :param inherited: A value indicating whether this layout node has been inherited. from a parent layout.  This is expected to only be only set by the combiner.
    :type inherited: bool
    :param is_contribution: A value indicating if the layout node is contribution or not.
    :type is_contribution: bool
    :param label: Label for the field.
    :type label: str
    :param metadata: Inner text of the control.
    :type metadata: str
    :param order: Order in which the control should appear in its group.
    :type order: int
    :param overridden: A value indicating whether this layout node has been overridden . by a child layout.
    :type overridden: bool
    :param read_only: A value indicating if the control is readonly.
    :type read_only: bool
    :param visible: A value indicating if the control should be hidden or not.
    :type visible: bool
    :param watermark: Watermark text for the textbox.
    :type watermark: str
    """

    _attribute_map = {
        'contribution': {'key': 'contribution', 'type': 'WitContribution'},
        'control_type': {'key': 'controlType', 'type': 'str'},
        'height': {'key': 'height', 'type': 'int'},
        'id': {'key': 'id', 'type': 'str'},
        'inherited': {'key': 'inherited', 'type': 'bool'},
        'is_contribution': {'key': 'isContribution', 'type': 'bool'},
        'label': {'key': 'label', 'type': 'str'},
        'metadata': {'key': 'metadata', 'type': 'str'},
        'order': {'key': 'order', 'type': 'int'},
        'overridden': {'key': 'overridden', 'type': 'bool'},
        'read_only': {'key': 'readOnly', 'type': 'bool'},
        'visible': {'key': 'visible', 'type': 'bool'},
        'watermark': {'key': 'watermark', 'type': 'str'}
    }

    def __init__(self, contribution=None, control_type=None, height=None, id=None, inherited=None, is_contribution=None, label=None, metadata=None, order=None, overridden=None, read_only=None, visible=None, watermark=None):
        super(Control, self).__init__()
        self.contribution = contribution
        self.control_type = control_type
        self.height = height
        self.id = id
        self.inherited = inherited
        self.is_contribution = is_contribution
        self.label = label
        self.metadata = metadata
        self.order = order
        self.overridden = overridden
        self.read_only = read_only
        self.visible = visible
        self.watermark = watermark


class CreateProcessModel(Model):
    """CreateProcessModel.

    :param description: Description of the process
    :type description: str
    :param name: Name of the process
    :type name: str
    :param parent_process_type_id: The ID of the parent process
    :type parent_process_type_id: str
    :param reference_name: Reference name of the process
    :type reference_name: str
    """

    _attribute_map = {
        'description': {'key': 'description', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'parent_process_type_id': {'key': 'parentProcessTypeId', 'type': 'str'},
        'reference_name': {'key': 'referenceName', 'type': 'str'}
    }

    def __init__(self, description=None, name=None, parent_process_type_id=None, reference_name=None):
        super(CreateProcessModel, self).__init__()
        self.description = description
        self.name = name
        self.parent_process_type_id = parent_process_type_id
        self.reference_name = reference_name


class CreateProcessRuleRequest(Model):
    """CreateProcessRuleRequest.

    :param actions: List of actions to take when the rule is triggered.
    :type actions: list of :class:`RuleAction <azure.devops.v5_0.work_item_tracking.models.RuleAction>`
    :param conditions: List of conditions when the rule should be triggered.
    :type conditions: list of :class:`RuleCondition <azure.devops.v5_0.work_item_tracking.models.RuleCondition>`
    :param is_disabled: Indicates if the rule is disabled.
    :type is_disabled: bool
    :param name: Name for the rule.
    :type name: str
    """

    _attribute_map = {
        'actions': {'key': 'actions', 'type': '[RuleAction]'},
        'conditions': {'key': 'conditions', 'type': '[RuleCondition]'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'}
    }

    def __init__(self, actions=None, conditions=None, is_disabled=None, name=None):
        super(CreateProcessRuleRequest, self).__init__()
        self.actions = actions
        self.conditions = conditions
        self.is_disabled = is_disabled
        self.name = name


class CreateProcessWorkItemTypeRequest(Model):
    """CreateProcessWorkItemTypeRequest.

    :param color: Color hexadecimal code to represent the work item type
    :type color: str
    :param description: Description of the work item type
    :type description: str
    :param icon: Icon to represent the work item type
    :type icon: str
    :param inherits_from: Parent work item type for work item type
    :type inherits_from: str
    :param is_disabled: True if the work item type need to be disabled
    :type is_disabled: bool
    :param name: Name of work item type
    :type name: str
    """

    _attribute_map = {
        'color': {'key': 'color', 'type': 'str'},
        'description': {'key': 'description', 'type': 'str'},
        'icon': {'key': 'icon', 'type': 'str'},
        'inherits_from': {'key': 'inheritsFrom', 'type': 'str'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'}
    }

    def __init__(self, color=None, description=None, icon=None, inherits_from=None, is_disabled=None, name=None):
        super(CreateProcessWorkItemTypeRequest, self).__init__()
        self.color = color
        self.description = description
        self.icon = icon
        self.inherits_from = inherits_from
        self.is_disabled = is_disabled
        self.name = name


class Extension(Model):
    """Extension.

    :param id: Id of the extension
    :type id: str
    """

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'}
    }

    def __init__(self, id=None):
        super(Extension, self).__init__()
        self.id = id


class FieldModel(Model):
    """FieldModel.

    :param description:
    :type description: str
    :param id:
    :type id: str
    :param is_identity:
    :type is_identity: bool
    :param name:
    :type name: str
    :param type:
    :type type: object
    :param url:
    :type url: str
    """

    _attribute_map = {
        'description': {'key': 'description', 'type': 'str'},
        'id': {'key': 'id', 'type': 'str'},
        'is_identity': {'key': 'isIdentity', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'},
        'type': {'key': 'type', 'type': 'object'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, description=None, id=None, is_identity=None, name=None, type=None, url=None):
        super(FieldModel, self).__init__()
        self.description = description
        self.id = id
        self.is_identity = is_identity
        self.name = name
        self.type = type
        self.url = url


class FieldRuleModel(Model):
    """FieldRuleModel.

    :param actions:
    :type actions: list of :class:`RuleActionModel <azure.devops.v5_0.work_item_tracking.models.RuleActionModel>`
    :param conditions:
    :type conditions: list of :class:`RuleConditionModel <azure.devops.v5_0.work_item_tracking.models.RuleConditionModel>`
    :param friendly_name:
    :type friendly_name: str
    :param id:
    :type id: str
    :param is_disabled:
    :type is_disabled: bool
    :param is_system:
    :type is_system: bool
    """

    _attribute_map = {
        'actions': {'key': 'actions', 'type': '[RuleActionModel]'},
        'conditions': {'key': 'conditions', 'type': '[RuleConditionModel]'},
        'friendly_name': {'key': 'friendlyName', 'type': 'str'},
        'id': {'key': 'id', 'type': 'str'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'},
        'is_system': {'key': 'isSystem', 'type': 'bool'}
    }

    def __init__(self, actions=None, conditions=None, friendly_name=None, id=None, is_disabled=None, is_system=None):
        super(FieldRuleModel, self).__init__()
        self.actions = actions
        self.conditions = conditions
        self.friendly_name = friendly_name
        self.id = id
        self.is_disabled = is_disabled
        self.is_system = is_system


class FormLayout(Model):
    """FormLayout.

    :param extensions: Gets and sets extensions list.
    :type extensions: list of :class:`Extension <azure.devops.v5_0.work_item_tracking.models.Extension>`
    :param pages: Top level tabs of the layout.
    :type pages: list of :class:`Page <azure.devops.v5_0.work_item_tracking.models.Page>`
    :param system_controls: Headers controls of the layout.
    :type system_controls: list of :class:`Control <azure.devops.v5_0.work_item_tracking.models.Control>`
    """

    _attribute_map = {
        'extensions': {'key': 'extensions', 'type': '[Extension]'},
        'pages': {'key': 'pages', 'type': '[Page]'},
        'system_controls': {'key': 'systemControls', 'type': '[Control]'}
    }

    def __init__(self, extensions=None, pages=None, system_controls=None):
        super(FormLayout, self).__init__()
        self.extensions = extensions
        self.pages = pages
        self.system_controls = system_controls


class Group(Model):
    """Group.

    :param contribution: Contribution for the group.
    :type contribution: :class:`WitContribution <azure.devops.v5_0.work_item_tracking.models.WitContribution>`
    :param controls: Controls to be put in the group.
    :type controls: list of :class:`Control <azure.devops.v5_0.work_item_tracking.models.Control>`
    :param height: The height for the contribution.
    :type height: int
    :param id: The id for the layout node.
    :type id: str
    :param inherited: A value indicating whether this layout node has been inherited from a parent layout.  This is expected to only be only set by the combiner.
    :type inherited: bool
    :param is_contribution: A value indicating if the layout node is contribution are not.
    :type is_contribution: bool
    :param label: Label for the group.
    :type label: str
    :param order: Order in which the group should appear in the section.
    :type order: int
    :param overridden: A value indicating whether this layout node has been overridden by a child layout.
    :type overridden: bool
    :param visible: A value indicating if the group should be hidden or not.
    :type visible: bool
    """

    _attribute_map = {
        'contribution': {'key': 'contribution', 'type': 'WitContribution'},
        'controls': {'key': 'controls', 'type': '[Control]'},
        'height': {'key': 'height', 'type': 'int'},
        'id': {'key': 'id', 'type': 'str'},
        'inherited': {'key': 'inherited', 'type': 'bool'},
        'is_contribution': {'key': 'isContribution', 'type': 'bool'},
        'label': {'key': 'label', 'type': 'str'},
        'order': {'key': 'order', 'type': 'int'},
        'overridden': {'key': 'overridden', 'type': 'bool'},
        'visible': {'key': 'visible', 'type': 'bool'}
    }

    def __init__(self, contribution=None, controls=None, height=None, id=None, inherited=None, is_contribution=None, label=None, order=None, overridden=None, visible=None):
        super(Group, self).__init__()
        self.contribution = contribution
        self.controls = controls
        self.height = height
        self.id = id
        self.inherited = inherited
        self.is_contribution = is_contribution
        self.label = label
        self.order = order
        self.overridden = overridden
        self.visible = visible


class HideStateModel(Model):
    """HideStateModel.

    :param hidden:
    :type hidden: bool
    """

    _attribute_map = {
        'hidden': {'key': 'hidden', 'type': 'bool'}
    }

    def __init__(self, hidden=None):
        super(HideStateModel, self).__init__()
        self.hidden = hidden


class Page(Model):
    """Page.

    :param contribution: Contribution for the page.
    :type contribution: :class:`WitContribution <azure.devops.v5_0.work_item_tracking.models.WitContribution>`
    :param id: The id for the layout node.
    :type id: str
    :param inherited: A value indicating whether this layout node has been inherited from a parent layout.  This is expected to only be only set by the combiner.
    :type inherited: bool
    :param is_contribution: A value indicating if the layout node is contribution are not.
    :type is_contribution: bool
    :param label: The label for the page.
    :type label: str
    :param locked: A value indicating whether any user operations are permitted on this page and the contents of this page
    :type locked: bool
    :param order: Order in which the page should appear in the layout.
    :type order: int
    :param overridden: A value indicating whether this layout node has been overridden by a child layout.
    :type overridden: bool
    :param page_type: The icon for the page.
    :type page_type: object
    :param sections: The sections of the page.
    :type sections: list of :class:`Section <azure.devops.v5_0.work_item_tracking.models.Section>`
    :param visible: A value indicating if the page should be hidden or not.
    :type visible: bool
    """

    _attribute_map = {
        'contribution': {'key': 'contribution', 'type': 'WitContribution'},
        'id': {'key': 'id', 'type': 'str'},
        'inherited': {'key': 'inherited', 'type': 'bool'},
        'is_contribution': {'key': 'isContribution', 'type': 'bool'},
        'label': {'key': 'label', 'type': 'str'},
        'locked': {'key': 'locked', 'type': 'bool'},
        'order': {'key': 'order', 'type': 'int'},
        'overridden': {'key': 'overridden', 'type': 'bool'},
        'page_type': {'key': 'pageType', 'type': 'object'},
        'sections': {'key': 'sections', 'type': '[Section]'},
        'visible': {'key': 'visible', 'type': 'bool'}
    }

    def __init__(self, contribution=None, id=None, inherited=None, is_contribution=None, label=None, locked=None, order=None, overridden=None, page_type=None, sections=None, visible=None):
        super(Page, self).__init__()
        self.contribution = contribution
        self.id = id
        self.inherited = inherited
        self.is_contribution = is_contribution
        self.label = label
        self.locked = locked
        self.order = order
        self.overridden = overridden
        self.page_type = page_type
        self.sections = sections
        self.visible = visible


class PickListMetadata(Model):
    """PickListMetadata.

    :param id: ID of the picklist
    :type id: str
    :param is_suggested: Indicates whether items outside of suggested list are allowed
    :type is_suggested: bool
    :param name: Name of the picklist
    :type name: str
    :param type: DataType of picklist
    :type type: str
    :param url: Url of the picklist
    :type url: str
    """

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'is_suggested': {'key': 'isSuggested', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'},
        'type': {'key': 'type', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, id=None, is_suggested=None, name=None, type=None, url=None):
        super(PickListMetadata, self).__init__()
        self.id = id
        self.is_suggested = is_suggested
        self.name = name
        self.type = type
        self.url = url


class ProcessBehavior(Model):
    """ProcessBehavior.

    :param color: Color.
    :type color: str
    :param customization: Indicates the type of customization on this work item. System behaviors are inherited from parent process but not modified. Inherited behaviors are modified modified behaviors that were inherited from parent process. Custom behaviors are behaviors created by user in current process.
    :type customization: object
    :param description: . Description
    :type description: str
    :param fields: Process Behavior Fields.
    :type fields: list of :class:`ProcessBehaviorField <azure.devops.v5_0.work_item_tracking.models.ProcessBehaviorField>`
    :param inherits: Parent behavior reference.
    :type inherits: :class:`ProcessBehaviorReference <azure.devops.v5_0.work_item_tracking.models.ProcessBehaviorReference>`
    :param name: Behavior Name.
    :type name: str
    :param rank: Rank of the behavior
    :type rank: int
    :param reference_name: Behavior Id
    :type reference_name: str
    :param url: Url of the behavior.
    :type url: str
    """

    _attribute_map = {
        'color': {'key': 'color', 'type': 'str'},
        'customization': {'key': 'customization', 'type': 'object'},
        'description': {'key': 'description', 'type': 'str'},
        'fields': {'key': 'fields', 'type': '[ProcessBehaviorField]'},
        'inherits': {'key': 'inherits', 'type': 'ProcessBehaviorReference'},
        'name': {'key': 'name', 'type': 'str'},
        'rank': {'key': 'rank', 'type': 'int'},
        'reference_name': {'key': 'referenceName', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, color=None, customization=None, description=None, fields=None, inherits=None, name=None, rank=None, reference_name=None, url=None):
        super(ProcessBehavior, self).__init__()
        self.color = color
        self.customization = customization
        self.description = description
        self.fields = fields
        self.inherits = inherits
        self.name = name
        self.rank = rank
        self.reference_name = reference_name
        self.url = url


class ProcessBehaviorCreateRequest(Model):
    """ProcessBehaviorCreateRequest.

    :param color: Color.
    :type color: str
    :param inherits: Parent behavior id.
    :type inherits: str
    :param name: Name of the behavior.
    :type name: str
    :param reference_name: ReferenceName is optional, if not specified will be auto-generated.
    :type reference_name: str
    """

    _attribute_map = {
        'color': {'key': 'color', 'type': 'str'},
        'inherits': {'key': 'inherits', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'reference_name': {'key': 'referenceName', 'type': 'str'}
    }

    def __init__(self, color=None, inherits=None, name=None, reference_name=None):
        super(ProcessBehaviorCreateRequest, self).__init__()
        self.color = color
        self.inherits = inherits
        self.name = name
        self.reference_name = reference_name


class ProcessBehaviorField(Model):
    """ProcessBehaviorField.

    :param name: Name of the field.
    :type name: str
    :param reference_name: Reference name of the field.
    :type reference_name: str
    :param url: Url to field.
    :type url: str
    """

    _attribute_map = {
        'name': {'key': 'name', 'type': 'str'},
        'reference_name': {'key': 'referenceName', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, name=None, reference_name=None, url=None):
        super(ProcessBehaviorField, self).__init__()
        self.name = name
        self.reference_name = reference_name
        self.url = url


class ProcessBehaviorReference(Model):
    """ProcessBehaviorReference.

    :param behavior_ref_name: Id of a Behavior.
    :type behavior_ref_name: str
    :param url: Url to behavior.
    :type url: str
    """

    _attribute_map = {
        'behavior_ref_name': {'key': 'behaviorRefName', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, behavior_ref_name=None, url=None):
        super(ProcessBehaviorReference, self).__init__()
        self.behavior_ref_name = behavior_ref_name
        self.url = url


class ProcessBehaviorUpdateRequest(Model):
    """ProcessBehaviorUpdateRequest.

    :param color: Color.
    :type color: str
    :param name: Behavior Name.
    :type name: str
    """

    _attribute_map = {
        'color': {'key': 'color', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'}
    }

    def __init__(self, color=None, name=None):
        super(ProcessBehaviorUpdateRequest, self).__init__()
        self.color = color
        self.name = name


class ProcessInfo(Model):
    """ProcessInfo.

    :param customization_type: Indicates the type of customization on this process. System Process is default process. Inherited Process is modified process that was System process before.
    :type customization_type: object
    :param description: Description of the process.
    :type description: str
    :param is_default: Is the process default.
    :type is_default: bool
    :param is_enabled: Is the process enabled.
    :type is_enabled: bool
    :param name: Name of the process.
    :type name: str
    :param parent_process_type_id: ID of the parent process.
    :type parent_process_type_id: str
    :param projects: Projects in this process to which the user is subscribed to.
    :type projects: list of :class:`ProjectReference <azure.devops.v5_0.work_item_tracking.models.ProjectReference>`
    :param reference_name: Reference name of the process.
    :type reference_name: str
    :param type_id: The ID of the process.
    :type type_id: str
    """

    _attribute_map = {
        'customization_type': {'key': 'customizationType', 'type': 'object'},
        'description': {'key': 'description', 'type': 'str'},
        'is_default': {'key': 'isDefault', 'type': 'bool'},
        'is_enabled': {'key': 'isEnabled', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'},
        'parent_process_type_id': {'key': 'parentProcessTypeId', 'type': 'str'},
        'projects': {'key': 'projects', 'type': '[ProjectReference]'},
        'reference_name': {'key': 'referenceName', 'type': 'str'},
        'type_id': {'key': 'typeId', 'type': 'str'}
    }

    def __init__(self, customization_type=None, description=None, is_default=None, is_enabled=None, name=None, parent_process_type_id=None, projects=None, reference_name=None, type_id=None):
        super(ProcessInfo, self).__init__()
        self.customization_type = customization_type
        self.description = description
        self.is_default = is_default
        self.is_enabled = is_enabled
        self.name = name
        self.parent_process_type_id = parent_process_type_id
        self.projects = projects
        self.reference_name = reference_name
        self.type_id = type_id


class ProcessModel(Model):
    """ProcessModel.

    :param description: Description of the process
    :type description: str
    :param name: Name of the process
    :type name: str
    :param projects: Projects in this process
    :type projects: list of :class:`ProjectReference <azure.devops.v5_0.work_item_tracking.models.ProjectReference>`
    :param properties: Properties of the process
    :type properties: :class:`ProcessProperties <azure.devops.v5_0.work_item_tracking.models.ProcessProperties>`
    :param reference_name: Reference name of the process
    :type reference_name: str
    :param type_id: The ID of the process
    :type type_id: str
    """

    _attribute_map = {
        'description': {'key': 'description', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'projects': {'key': 'projects', 'type': '[ProjectReference]'},
        'properties': {'key': 'properties', 'type': 'ProcessProperties'},
        'reference_name': {'key': 'referenceName', 'type': 'str'},
        'type_id': {'key': 'typeId', 'type': 'str'}
    }

    def __init__(self, description=None, name=None, projects=None, properties=None, reference_name=None, type_id=None):
        super(ProcessModel, self).__init__()
        self.description = description
        self.name = name
        self.projects = projects
        self.properties = properties
        self.reference_name = reference_name
        self.type_id = type_id


class ProcessProperties(Model):
    """ProcessProperties.

    :param class_: Class of the process.
    :type class_: object
    :param is_default: Is the process default process.
    :type is_default: bool
    :param is_enabled: Is the process enabled.
    :type is_enabled: bool
    :param parent_process_type_id: ID of the parent process.
    :type parent_process_type_id: str
    :param version: Version of the process.
    :type version: str
    """

    _attribute_map = {
        'class_': {'key': 'class', 'type': 'object'},
        'is_default': {'key': 'isDefault', 'type': 'bool'},
        'is_enabled': {'key': 'isEnabled', 'type': 'bool'},
        'parent_process_type_id': {'key': 'parentProcessTypeId', 'type': 'str'},
        'version': {'key': 'version', 'type': 'str'}
    }

    def __init__(self, class_=None, is_default=None, is_enabled=None, parent_process_type_id=None, version=None):
        super(ProcessProperties, self).__init__()
        self.class_ = class_
        self.is_default = is_default
        self.is_enabled = is_enabled
        self.parent_process_type_id = parent_process_type_id
        self.version = version


class ProcessRule(CreateProcessRuleRequest):
    """ProcessRule.

    :param actions: List of actions to take when the rule is triggered.
    :type actions: list of :class:`RuleAction <azure.devops.v5_0.work_item_tracking.models.RuleAction>`
    :param conditions: List of conditions when the rule should be triggered.
    :type conditions: list of :class:`RuleCondition <azure.devops.v5_0.work_item_tracking.models.RuleCondition>`
    :param is_disabled: Indicates if the rule is disabled.
    :type is_disabled: bool
    :param name: Name for the rule.
    :type name: str
    :param customization_type: Indicates if the rule is system generated or created by user.
    :type customization_type: object
    :param id: Id to uniquely identify the rule.
    :type id: str
    :param url: Resource Url.
    :type url: str
    """

    _attribute_map = {
        'actions': {'key': 'actions', 'type': '[RuleAction]'},
        'conditions': {'key': 'conditions', 'type': '[RuleCondition]'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'},
        'customization_type': {'key': 'customizationType', 'type': 'object'},
        'id': {'key': 'id', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, actions=None, conditions=None, is_disabled=None, name=None, customization_type=None, id=None, url=None):
        super(ProcessRule, self).__init__(actions=actions, conditions=conditions, is_disabled=is_disabled, name=name)
        self.customization_type = customization_type
        self.id = id
        self.url = url


class ProcessWorkItemType(Model):
    """ProcessWorkItemType.

    :param behaviors:
    :type behaviors: list of :class:`WorkItemTypeBehavior <azure.devops.v5_0.work_item_tracking.models.WorkItemTypeBehavior>`
    :param color: Color hexadecimal code to represent the work item type
    :type color: str
    :param customization: Indicates the type of customization on this work item System work item types are inherited from parent process but not modified Inherited work item types are modified work item that were inherited from parent process Custom work item types are work item types that were created in the current process
    :type customization: object
    :param description: Description of the work item type
    :type description: str
    :param icon: Icon to represent the work item typ
    :type icon: str
    :param inherits: Reference name of the parent work item type
    :type inherits: str
    :param is_disabled: Indicates if a work item type is disabled
    :type is_disabled: bool
    :param layout:
    :type layout: :class:`FormLayout <azure.devops.v5_0.work_item_tracking.models.FormLayout>`
    :param name: Name of the work item type
    :type name: str
    :param reference_name: Reference name of work item type
    :type reference_name: str
    :param states:
    :type states: list of :class:`WorkItemStateResultModel <azure.devops.v5_0.work_item_tracking.models.WorkItemStateResultModel>`
    :param url: Url of the work item type
    :type url: str
    """

    _attribute_map = {
        'behaviors': {'key': 'behaviors', 'type': '[WorkItemTypeBehavior]'},
        'color': {'key': 'color', 'type': 'str'},
        'customization': {'key': 'customization', 'type': 'object'},
        'description': {'key': 'description', 'type': 'str'},
        'icon': {'key': 'icon', 'type': 'str'},
        'inherits': {'key': 'inherits', 'type': 'str'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'},
        'layout': {'key': 'layout', 'type': 'FormLayout'},
        'name': {'key': 'name', 'type': 'str'},
        'reference_name': {'key': 'referenceName', 'type': 'str'},
        'states': {'key': 'states', 'type': '[WorkItemStateResultModel]'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, behaviors=None, color=None, customization=None, description=None, icon=None, inherits=None, is_disabled=None, layout=None, name=None, reference_name=None, states=None, url=None):
        super(ProcessWorkItemType, self).__init__()
        self.behaviors = behaviors
        self.color = color
        self.customization = customization
        self.description = description
        self.icon = icon
        self.inherits = inherits
        self.is_disabled = is_disabled
        self.layout = layout
        self.name = name
        self.reference_name = reference_name
        self.states = states
        self.url = url


class ProcessWorkItemTypeField(Model):
    """ProcessWorkItemTypeField.

    :param allow_groups: Allow setting field value to a group identity. Only applies to identity fields.
    :type allow_groups: bool
    :param customization:
    :type customization: object
    :param default_value: The default value of the field.
    :type default_value: object
    :param description: Description of the field.
    :type description: str
    :param name: Name of the field.
    :type name: str
    :param read_only: If true the field cannot be edited.
    :type read_only: bool
    :param reference_name: Reference name of the field.
    :type reference_name: str
    :param required: If true the field cannot be empty.
    :type required: bool
    :param type: Type of the field.
    :type type: object
    :param url: Resource URL of the field.
    :type url: str
    """

    _attribute_map = {
        'allow_groups': {'key': 'allowGroups', 'type': 'bool'},
        'customization': {'key': 'customization', 'type': 'object'},
        'default_value': {'key': 'defaultValue', 'type': 'object'},
        'description': {'key': 'description', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'read_only': {'key': 'readOnly', 'type': 'bool'},
        'reference_name': {'key': 'referenceName', 'type': 'str'},
        'required': {'key': 'required', 'type': 'bool'},
        'type': {'key': 'type', 'type': 'object'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, allow_groups=None, customization=None, default_value=None, description=None, name=None, read_only=None, reference_name=None, required=None, type=None, url=None):
        super(ProcessWorkItemTypeField, self).__init__()
        self.allow_groups = allow_groups
        self.customization = customization
        self.default_value = default_value
        self.description = description
        self.name = name
        self.read_only = read_only
        self.reference_name = reference_name
        self.required = required
        self.type = type
        self.url = url


class ProjectReference(Model):
    """ProjectReference.

    :param description: Description of the project
    :type description: str
    :param id: The ID of the project
    :type id: str
    :param name: Name of the project
    :type name: str
    :param url: Url of the project
    :type url: str
    """

    _attribute_map = {
        'description': {'key': 'description', 'type': 'str'},
        'id': {'key': 'id', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, description=None, id=None, name=None, url=None):
        super(ProjectReference, self).__init__()
        self.description = description
        self.id = id
        self.name = name
        self.url = url


class RuleAction(Model):
    """RuleAction.

    :param action_type: Type of action to take when the rule is triggered.
    :type action_type: object
    :param target_field: Field on which the action should be taken.
    :type target_field: str
    :param value: Value to apply on target field, once the action is taken.
    :type value: str
    """

    _attribute_map = {
        'action_type': {'key': 'actionType', 'type': 'object'},
        'target_field': {'key': 'targetField', 'type': 'str'},
        'value': {'key': 'value', 'type': 'str'}
    }

    def __init__(self, action_type=None, target_field=None, value=None):
        super(RuleAction, self).__init__()
        self.action_type = action_type
        self.target_field = target_field
        self.value = value


class RuleActionModel(Model):
    """RuleActionModel.

    :param action_type:
    :type action_type: str
    :param target_field:
    :type target_field: str
    :param value:
    :type value: str
    """

    _attribute_map = {
        'action_type': {'key': 'actionType', 'type': 'str'},
        'target_field': {'key': 'targetField', 'type': 'str'},
        'value': {'key': 'value', 'type': 'str'}
    }

    def __init__(self, action_type=None, target_field=None, value=None):
        super(RuleActionModel, self).__init__()
        self.action_type = action_type
        self.target_field = target_field
        self.value = value


class RuleCondition(Model):
    """RuleCondition.

    :param condition_type: Type of condition. $When. This condition limits the execution of its children to cases when another field has a particular value, i.e. when the Is value of the referenced field is equal to the given literal value. $WhenNot.This condition limits the execution of its children to cases when another field does not have a particular value, i.e.when the Is value of the referenced field is not equal to the given literal value. $WhenChanged.This condition limits the execution of its children to cases when another field has changed, i.e.when the Is value of the referenced field is not equal to the Was value of that field. $WhenNotChanged.This condition limits the execution of its children to cases when another field has not changed, i.e.when the Is value of the referenced field is equal to the Was value of that field.
    :type condition_type: object
    :param field: Field that defines condition.
    :type field: str
    :param value: Value of field to define the condition for rule.
    :type value: str
    """

    _attribute_map = {
        'condition_type': {'key': 'conditionType', 'type': 'object'},
        'field': {'key': 'field', 'type': 'str'},
        'value': {'key': 'value', 'type': 'str'}
    }

    def __init__(self, condition_type=None, field=None, value=None):
        super(RuleCondition, self).__init__()
        self.condition_type = condition_type
        self.field = field
        self.value = value


class RuleConditionModel(Model):
    """RuleConditionModel.

    :param condition_type:
    :type condition_type: str
    :param field:
    :type field: str
    :param value:
    :type value: str
    """

    _attribute_map = {
        'condition_type': {'key': 'conditionType', 'type': 'str'},
        'field': {'key': 'field', 'type': 'str'},
        'value': {'key': 'value', 'type': 'str'}
    }

    def __init__(self, condition_type=None, field=None, value=None):
        super(RuleConditionModel, self).__init__()
        self.condition_type = condition_type
        self.field = field
        self.value = value


class Section(Model):
    """Section.

    :param groups: List of child groups in this section
    :type groups: list of :class:`Group <azure.devops.v5_0.work_item_tracking.models.Group>`
    :param id: The id for the layout node.
    :type id: str
    :param overridden: A value indicating whether this layout node has been overridden by a child layout.
    :type overridden: bool
    """

    _attribute_map = {
        'groups': {'key': 'groups', 'type': '[Group]'},
        'id': {'key': 'id', 'type': 'str'},
        'overridden': {'key': 'overridden', 'type': 'bool'}
    }

    def __init__(self, groups=None, id=None, overridden=None):
        super(Section, self).__init__()
        self.groups = groups
        self.id = id
        self.overridden = overridden


class UpdateProcessModel(Model):
    """UpdateProcessModel.

    :param description: New description of the process
    :type description: str
    :param is_default: If true new projects will use this process by default
    :type is_default: bool
    :param is_enabled: If false the process will be disabled and cannot be used to create projects
    :type is_enabled: bool
    :param name: New name of the process
    :type name: str
    """

    _attribute_map = {
        'description': {'key': 'description', 'type': 'str'},
        'is_default': {'key': 'isDefault', 'type': 'bool'},
        'is_enabled': {'key': 'isEnabled', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'}
    }

    def __init__(self, description=None, is_default=None, is_enabled=None, name=None):
        super(UpdateProcessModel, self).__init__()
        self.description = description
        self.is_default = is_default
        self.is_enabled = is_enabled
        self.name = name


class UpdateProcessRuleRequest(CreateProcessRuleRequest):
    """UpdateProcessRuleRequest.

    :param actions: List of actions to take when the rule is triggered.
    :type actions: list of :class:`RuleAction <azure.devops.v5_0.work_item_tracking.models.RuleAction>`
    :param conditions: List of conditions when the rule should be triggered.
    :type conditions: list of :class:`RuleCondition <azure.devops.v5_0.work_item_tracking.models.RuleCondition>`
    :param is_disabled: Indicates if the rule is disabled.
    :type is_disabled: bool
    :param name: Name for the rule.
    :type name: str
    :param id: Id to uniquely identify the rule.
    :type id: str
    """

    _attribute_map = {
        'actions': {'key': 'actions', 'type': '[RuleAction]'},
        'conditions': {'key': 'conditions', 'type': '[RuleCondition]'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'},
        'id': {'key': 'id', 'type': 'str'}
    }

    def __init__(self, actions=None, conditions=None, is_disabled=None, name=None, id=None):
        super(UpdateProcessRuleRequest, self).__init__(actions=actions, conditions=conditions, is_disabled=is_disabled, name=name)
        self.id = id


class UpdateProcessWorkItemTypeFieldRequest(Model):
    """UpdateProcessWorkItemTypeFieldRequest.

    :param allow_groups: Allow setting field value to a group identity. Only applies to identity fields.
    :type allow_groups: bool
    :param default_value: The default value of the field.
    :type default_value: object
    :param read_only: If true the field cannot be edited.
    :type read_only: bool
    :param required: The default value of the field.
    :type required: bool
    """

    _attribute_map = {
        'allow_groups': {'key': 'allowGroups', 'type': 'bool'},
        'default_value': {'key': 'defaultValue', 'type': 'object'},
        'read_only': {'key': 'readOnly', 'type': 'bool'},
        'required': {'key': 'required', 'type': 'bool'}
    }

    def __init__(self, allow_groups=None, default_value=None, read_only=None, required=None):
        super(UpdateProcessWorkItemTypeFieldRequest, self).__init__()
        self.allow_groups = allow_groups
        self.default_value = default_value
        self.read_only = read_only
        self.required = required


class UpdateProcessWorkItemTypeRequest(Model):
    """UpdateProcessWorkItemTypeRequest.

    :param color: Color of the work item type
    :type color: str
    :param description: Description of the work item type
    :type description: str
    :param icon: Icon of the work item type
    :type icon: str
    :param is_disabled: If set will disable the work item type
    :type is_disabled: bool
    """

    _attribute_map = {
        'color': {'key': 'color', 'type': 'str'},
        'description': {'key': 'description', 'type': 'str'},
        'icon': {'key': 'icon', 'type': 'str'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'}
    }

    def __init__(self, color=None, description=None, icon=None, is_disabled=None):
        super(UpdateProcessWorkItemTypeRequest, self).__init__()
        self.color = color
        self.description = description
        self.icon = icon
        self.is_disabled = is_disabled


class WitContribution(Model):
    """WitContribution.

    :param contribution_id: The id for the contribution.
    :type contribution_id: str
    :param height: The height for the contribution.
    :type height: int
    :param inputs: A dictionary holding key value pairs for contribution inputs.
    :type inputs: dict
    :param show_on_deleted_work_item: A value indicating if the contribution should be show on deleted workItem.
    :type show_on_deleted_work_item: bool
    """

    _attribute_map = {
        'contribution_id': {'key': 'contributionId', 'type': 'str'},
        'height': {'key': 'height', 'type': 'int'},
        'inputs': {'key': 'inputs', 'type': '{object}'},
        'show_on_deleted_work_item': {'key': 'showOnDeletedWorkItem', 'type': 'bool'}
    }

    def __init__(self, contribution_id=None, height=None, inputs=None, show_on_deleted_work_item=None):
        super(WitContribution, self).__init__()
        self.contribution_id = contribution_id
        self.height = height
        self.inputs = inputs
        self.show_on_deleted_work_item = show_on_deleted_work_item


class WorkItemBehavior(Model):
    """WorkItemBehavior.

    :param abstract:
    :type abstract: bool
    :param color:
    :type color: str
    :param description:
    :type description: str
    :param fields:
    :type fields: list of :class:`WorkItemBehaviorField <azure.devops.v5_0.work_item_tracking.models.WorkItemBehaviorField>`
    :param id:
    :type id: str
    :param inherits:
    :type inherits: :class:`WorkItemBehaviorReference <azure.devops.v5_0.work_item_tracking.models.WorkItemBehaviorReference>`
    :param name:
    :type name: str
    :param overriden:
    :type overriden: bool
    :param rank:
    :type rank: int
    :param url:
    :type url: str
    """

    _attribute_map = {
        'abstract': {'key': 'abstract', 'type': 'bool'},
        'color': {'key': 'color', 'type': 'str'},
        'description': {'key': 'description', 'type': 'str'},
        'fields': {'key': 'fields', 'type': '[WorkItemBehaviorField]'},
        'id': {'key': 'id', 'type': 'str'},
        'inherits': {'key': 'inherits', 'type': 'WorkItemBehaviorReference'},
        'name': {'key': 'name', 'type': 'str'},
        'overriden': {'key': 'overriden', 'type': 'bool'},
        'rank': {'key': 'rank', 'type': 'int'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, abstract=None, color=None, description=None, fields=None, id=None, inherits=None, name=None, overriden=None, rank=None, url=None):
        super(WorkItemBehavior, self).__init__()
        self.abstract = abstract
        self.color = color
        self.description = description
        self.fields = fields
        self.id = id
        self.inherits = inherits
        self.name = name
        self.overriden = overriden
        self.rank = rank
        self.url = url


class WorkItemBehaviorField(Model):
    """WorkItemBehaviorField.

    :param behavior_field_id:
    :type behavior_field_id: str
    :param id:
    :type id: str
    :param url:
    :type url: str
    """

    _attribute_map = {
        'behavior_field_id': {'key': 'behaviorFieldId', 'type': 'str'},
        'id': {'key': 'id', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, behavior_field_id=None, id=None, url=None):
        super(WorkItemBehaviorField, self).__init__()
        self.behavior_field_id = behavior_field_id
        self.id = id
        self.url = url


class WorkItemBehaviorReference(Model):
    """WorkItemBehaviorReference.

    :param id:
    :type id: str
    :param url:
    :type url: str
    """

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, id=None, url=None):
        super(WorkItemBehaviorReference, self).__init__()
        self.id = id
        self.url = url


class WorkItemStateInputModel(Model):
    """WorkItemStateInputModel.

    :param color: Color of the state
    :type color: str
    :param name: Name of the state
    :type name: str
    :param order: Order in which state should appear
    :type order: int
    :param state_category: Category of the state
    :type state_category: str
    """

    _attribute_map = {
        'color': {'key': 'color', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'order': {'key': 'order', 'type': 'int'},
        'state_category': {'key': 'stateCategory', 'type': 'str'}
    }

    def __init__(self, color=None, name=None, order=None, state_category=None):
        super(WorkItemStateInputModel, self).__init__()
        self.color = color
        self.name = name
        self.order = order
        self.state_category = state_category


class WorkItemStateResultModel(Model):
    """WorkItemStateResultModel.

    :param color:
    :type color: str
    :param customization_type:
    :type customization_type: object
    :param hidden:
    :type hidden: bool
    :param id:
    :type id: str
    :param name:
    :type name: str
    :param order:
    :type order: int
    :param state_category:
    :type state_category: str
    :param url:
    :type url: str
    """

    _attribute_map = {
        'color': {'key': 'color', 'type': 'str'},
        'customization_type': {'key': 'customizationType', 'type': 'object'},
        'hidden': {'key': 'hidden', 'type': 'bool'},
        'id': {'key': 'id', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'order': {'key': 'order', 'type': 'int'},
        'state_category': {'key': 'stateCategory', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, color=None, customization_type=None, hidden=None, id=None, name=None, order=None, state_category=None, url=None):
        super(WorkItemStateResultModel, self).__init__()
        self.color = color
        self.customization_type = customization_type
        self.hidden = hidden
        self.id = id
        self.name = name
        self.order = order
        self.state_category = state_category
        self.url = url


class WorkItemTypeBehavior(Model):
    """WorkItemTypeBehavior.

    :param behavior: Reference to the behavior of a work item type
    :type behavior: :class:`WorkItemBehaviorReference <azure.devops.v5_0.work_item_tracking.models.WorkItemBehaviorReference>`
    :param is_default: If true the work item type is the default work item type in the behavior
    :type is_default: bool
    :param url: URL of the work item type behavior
    :type url: str
    """

    _attribute_map = {
        'behavior': {'key': 'behavior', 'type': 'WorkItemBehaviorReference'},
        'is_default': {'key': 'isDefault', 'type': 'bool'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, behavior=None, is_default=None, url=None):
        super(WorkItemTypeBehavior, self).__init__()
        self.behavior = behavior
        self.is_default = is_default
        self.url = url


class WorkItemTypeModel(Model):
    """WorkItemTypeModel.

    :param behaviors:
    :type behaviors: list of :class:`WorkItemTypeBehavior <azure.devops.v5_0.work_item_tracking.models.WorkItemTypeBehavior>`
    :param class_:
    :type class_: object
    :param color:
    :type color: str
    :param description:
    :type description: str
    :param icon:
    :type icon: str
    :param id:
    :type id: str
    :param inherits: Parent WIT Id/Internal ReferenceName that it inherits from
    :type inherits: str
    :param is_disabled:
    :type is_disabled: bool
    :param layout:
    :type layout: :class:`FormLayout <azure.devops.v5_0.work_item_tracking.models.FormLayout>`
    :param name:
    :type name: str
    :param states:
    :type states: list of :class:`WorkItemStateResultModel <azure.devops.v5_0.work_item_tracking.models.WorkItemStateResultModel>`
    :param url:
    :type url: str
    """

    _attribute_map = {
        'behaviors': {'key': 'behaviors', 'type': '[WorkItemTypeBehavior]'},
        'class_': {'key': 'class', 'type': 'object'},
        'color': {'key': 'color', 'type': 'str'},
        'description': {'key': 'description', 'type': 'str'},
        'icon': {'key': 'icon', 'type': 'str'},
        'id': {'key': 'id', 'type': 'str'},
        'inherits': {'key': 'inherits', 'type': 'str'},
        'is_disabled': {'key': 'isDisabled', 'type': 'bool'},
        'layout': {'key': 'layout', 'type': 'FormLayout'},
        'name': {'key': 'name', 'type': 'str'},
        'states': {'key': 'states', 'type': '[WorkItemStateResultModel]'},
        'url': {'key': 'url', 'type': 'str'}
    }

    def __init__(self, behaviors=None, class_=None, color=None, description=None, icon=None, id=None, inherits=None, is_disabled=None, layout=None, name=None, states=None, url=None):
        super(WorkItemTypeModel, self).__init__()
        self.behaviors = behaviors
        self.class_ = class_
        self.color = color
        self.description = description
        self.icon = icon
        self.id = id
        self.inherits = inherits
        self.is_disabled = is_disabled
        self.layout = layout
        self.name = name
        self.states = states
        self.url = url


class PickList(PickListMetadata):
    """PickList.

    :param id: ID of the picklist
    :type id: str
    :param is_suggested: Indicates whether items outside of suggested list are allowed
    :type is_suggested: bool
    :param name: Name of the picklist
    :type name: str
    :param type: DataType of picklist
    :type type: str
    :param url: Url of the picklist
    :type url: str
    :param items: A list of PicklistItemModel.
    :type items: list of str
    """

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'is_suggested': {'key': 'isSuggested', 'type': 'bool'},
        'name': {'key': 'name', 'type': 'str'},
        'type': {'key': 'type', 'type': 'str'},
        'url': {'key': 'url', 'type': 'str'},
        'items': {'key': 'items', 'type': '[str]'}
    }

    def __init__(self, id=None, is_suggested=None, name=None, type=None, url=None, items=None):
        super(PickList, self).__init__(id=id, is_suggested=is_suggested, name=name, type=type, url=url)
        self.items = items


__all__ = [
    'AddProcessWorkItemTypeFieldRequest',
    'Control',
    'CreateProcessModel',
    'CreateProcessRuleRequest',
    'CreateProcessWorkItemTypeRequest',
    'Extension',
    'FieldModel',
    'FieldRuleModel',
    'FormLayout',
    'Group',
    'HideStateModel',
    'Page',
    'PickListMetadata',
    'ProcessBehavior',
    'ProcessBehaviorCreateRequest',
    'ProcessBehaviorField',
    'ProcessBehaviorReference',
    'ProcessBehaviorUpdateRequest',
    'ProcessInfo',
    'ProcessModel',
    'ProcessProperties',
    'ProcessRule',
    'ProcessWorkItemType',
    'ProcessWorkItemTypeField',
    'ProjectReference',
    'RuleAction',
    'RuleActionModel',
    'RuleCondition',
    'RuleConditionModel',
    'Section',
    'UpdateProcessModel',
    'UpdateProcessRuleRequest',
    'UpdateProcessWorkItemTypeFieldRequest',
    'UpdateProcessWorkItemTypeRequest',
    'WitContribution',
    'WorkItemBehavior',
    'WorkItemBehaviorField',
    'WorkItemBehaviorReference',
    'WorkItemStateInputModel',
    'WorkItemStateResultModel',
    'WorkItemTypeBehavior',
    'WorkItemTypeModel',
    'PickList',
]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os

from azure.cli.testsdk.scenario_tests import AllowLargeResponse
from .utilities.helper import (DevopsScenarioTest,
                               disable_telemetry,
                               get_random_name,
                               set_authentication,
                               get_test_org_from_env_variable)


DEVOPS_CLI_TEST_ORGANIZATION = get_test_org_from_env_variable() or 'Https://dev.azure.com/azuredevopsclitest'


class ReposRefTests(DevopsScenarioTest):

    @AllowLargeResponse(size_kb=3072)
    @disable_telemetry
    @set_authentication
    def test_ref_createListUpdateDelete(self):
        random_project_name = self.create_random_name(prefix='refsTests', length=15)
        random_repo_name = self.create_random_name(prefix='refsTests', length=15)
        self.cmd('az devops configure --defaults organization=' +  DEVOPS_CLI_TEST_ORGANIZATION + ' project=' + random_project_name)

        created_project_id = None

        try:
            create_project_command = 'az devops project create --name ' + random_project_name + ' --output json --detect false'
            project_create_output = self.cmd(create_project_command).get_output_in_json()
            created_project_id = project_create_output["id"]

            create_repo_command = 'az repos create --name ' + random_repo_name + ' -p ' +  created_project_id + ' --output json --detect false'
            repo_create_output = self.cmd(create_repo_command).get_output_in_json()
            create_repo_id = repo_create_output["id"]

            import_repo_command = 'az repos import create --git-url https://github.com/hkasera/snakes-and-ladders.git' + ' -p ' + created_project_id + ' -r ' + create_repo_id + ' --output json --detect false'
            import_repo_output = self.cmd(import_repo_command)

            REPO_NAME = '--repository {random_repo_name} --output json --detect false'.format(random_repo_name=random_repo_name)
            REF_NAME = 'heads/branchnametocreate'

            list_command = 'az repos ref list {}'.format(REPO_NAME)
            list_refs = self.cmd(list_command).get_output_in_json()

            refs_nbre = len(list_refs)
            assert refs_nbre > 0

            master_object_id = list_refs[0]['objectId']
            assert master_object_id is not None

            # create a new reference
            create_command = 'az repos ref create --name {} --object-id {} {}'.format(REF_NAME, master_object_id, REPO_NAME)
            created_ref = self.cmd(create_command).get_output_in_json()
            created_object_id = created_ref['newObjectId']
            assert created_object_id is not None
            assert created_ref['updateStatus'] == 'succeeded'
            assert created_ref['success'] is True

            # lock the reference
            lock_command = 'az repos ref lock --name {} {}'.format(REF_NAME, REPO_NAME)
            locked_ref = self.cmd(lock_command).get_output_in_json()
            assert locked_ref['isLocked'] is True
            assert locked_ref['isLockedBy'] is not None

            # unlock the reference
            unlock_command = 'az repos ref unlock --name {} {}'.format(REF_NAME, REPO_NAME)
            unlocked_ref = self.cmd(unlock_command).get_output_in_json()
            assert unlocked_ref['isLocked'] is None
            assert unlocked_ref['isLockedBy'] is None

            # delete the reference
            delete_command = 'az repos ref delete --name {} --object-id {} {}'.format(REF_NAME, created_object_id, REPO_NAME)
            deleted_ref = self.cmd(delete_command).get_output_in_json()
            assert deleted_ref is not None
            assert created_ref['updateStatus'] == 'succeeded'
            assert created_ref['success'] is True

        finally:
            if created_project_id is not None:
                delete_project_command = 'az devops project delete --id ' + created_project_id + ' --output json --detect false -y'
                self.cmd(delete_project_command)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# Generated file, DO NOT EDIT
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------------------------

from ...v5_1.work_item_tracking.models import *
from .work_item_tracking_client import WorkItemTrackingClient

__all__ = [
    'AccountMyWorkResult',
    'AccountRecentActivityWorkItemModel',
    'AccountRecentActivityWorkItemModel2',
    'AccountRecentActivityWorkItemModelBase',
    'AccountRecentMentionWorkItemModel',
    'AccountWorkWorkItemModel',
    'ArtifactUriQuery',
    'ArtifactUriQueryResult',
    'AttachmentReference',
    'Comment',
    'CommentCreate',
    'CommentList',
    'CommentMention',
    'CommentReaction',
    'CommentUpdate',
    'CommentVersion',
    'FieldDependentRule',
    'FieldsToEvaluate',
    'GraphSubjectBase',
    'IdentityRef',
    'IdentityReference',
    'JsonPatchOperation',
    'Link',
    'ProjectWorkItemStateColors',
    'ProvisioningResult',
    'QueryBatchGetRequest',
    'QueryHierarchyItem',
    'QueryHierarchyItemsResult',
    'ReferenceLinks',
    'ReportingWorkItemLinksBatch',
    'ReportingWorkItemRevisionsBatch',
    'ReportingWorkItemRevisionsFilter',
    'StreamedBatch',
    'TeamContext',
    'Wiql',
    'WorkArtifactLink',
    'WorkItem',
    'WorkItemBatchGetRequest',
    'WorkItemClassificationNode',
    'WorkItemComment',
    'WorkItemComments',
    'WorkItemCommentVersionRef',
    'WorkItemDelete',
    'WorkItemDeleteReference',
    'WorkItemDeleteShallowReference',
    'WorkItemDeleteUpdate',
    'WorkItemField',
    'WorkItemFieldOperation',
    'WorkItemFieldReference',
    'WorkItemFieldUpdate',
    'WorkItemHistory',
    'WorkItemIcon',
    'WorkItemLink',
    'WorkItemNextStateOnTransition',
    'WorkItemQueryClause',
    'WorkItemQueryResult',
    'WorkItemQuerySortColumn',
    'WorkItemReference',
    'WorkItemRelation',
    'WorkItemRelationType',
    'WorkItemRelationUpdates',
    'WorkItemStateColor',
    'WorkItemStateTransition',
    'WorkItemTemplate',
    'WorkItemTemplateReference',
    'WorkItemTrackingReference',
    'WorkItemTrackingResource',
    'WorkItemTrackingResourceReference',
    'WorkItemType',
    'WorkItemTypeCategory',
    'WorkItemTypeColor',
    'WorkItemTypeColorAndIcon',
    'WorkItemTypeFieldInstance',
    'WorkItemTypeFieldInstanceBase',
    'WorkItemTypeFieldWithReferences',
    'WorkItemTypeReference',
    'WorkItemTypeStateColors',
    'WorkItemTypeTemplate',
    'WorkItemTypeTemplateUpdateModel',
    'WorkItemUpdate',
    'WorkItemTrackingClient'
]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest

try:
    # Attempt to load mock (works on Python 3.3 and above)
    from unittest.mock import patch
except ImportError:
    # Attempt to load mock (works on Python version below 3.3)
    from mock import patch

from azext_devops.dev.common.services import clear_connection_cache
from azext_devops.tests.utils.authentication import AuthenticatedTests
from azext_devops.dev.pipelines.pipeline_folders import (pipeline_folder_create,
                                                         pipeline_folder_delete,
                                                         pipeline_folder_list,
                                                         pipeline_folder_update)
from azext_devops.devops_sdk.v5_0.build.build_client import BuildClient
from azext_devops.devops_sdk.v5_0.build.models import Folder


class TestPipelinesFoldersMethods(AuthenticatedTests):

    _TEST_DEVOPS_ORGANIZATION = 'https://someorganization.visualstudio.com'
    _TEST_DEVOPS_PROJECT = 'MyProject'

    def setUp(self):
        self.authentication_setup()
        self.authenticate()

        self.get_client_patcher = patch('azext_devops.devops_sdk.connection.Connection.get_client')
        #start the patchers
        self.mock_get_client = self.get_client_patcher.start()
        # Set return values which will be same across tests
        self.mock_get_client.return_value = BuildClient(base_url=self._TEST_DEVOPS_ORGANIZATION)
        #clear connection cache before running each test
        clear_connection_cache()

    def tearDown(self):
        patch.stopall()

    def test_folder_create(self):
        with patch('azext_devops.devops_sdk.v5_0.build.build_client.BuildClient.create_folder') as mock_create_folder:
            # Creating folder
            new_folder = pipeline_folder_create(path='test', description='test description',
                           project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
            folder = Folder()
            folder.description = 'test description'
            folder.path = 'test'
            # assert
            mock_create_folder.assert_called_once_with(
                folder=folder, path='test', project=self._TEST_DEVOPS_PROJECT)

    def test_folder_delete(self):
        with patch('azext_devops.devops_sdk.v5_0.build.build_client.BuildClient.delete_folder') as mock_delete_folder:
            # deleting folder
            pipeline_folder_delete(
                path='test', project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_delete_folder.assert_called_once_with(
                path='test', project=self._TEST_DEVOPS_PROJECT)
    
    def test_folder_list(self):
        with patch('azext_devops.devops_sdk.v5_0.build.build_client.BuildClient.get_folders') as mock_list_folders:
            # listing folders
            folders = pipeline_folder_list(
                path='test', query_order='asc', project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
            # assert
            mock_list_folders.assert_called_once_with(
                path='test', query_order='folderAscending', project=self._TEST_DEVOPS_PROJECT)

    def test_folder_update(self):
        with patch('azext_devops.devops_sdk.v5_0.build.build_client.BuildClient.update_folder') as mock_update_folders:
            with patch('azext_devops.devops_sdk.v5_0.build.build_client.BuildClient.get_folders') as mock_list_folders:
                folder = Folder()
                folder.description ='hello world'
                folder.path = 'test'
                mock_list_folders.return_value = [folder]

                update_folder = Folder()
                update_folder.description ='hello world updated'
                update_folder.path = 'test123'

                updated_folder = pipeline_folder_update(
                    path='test', new_path='test123', new_description='hello world updated',
                    project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)

                # assert
                mock_list_folders.assert_called_once_with(
                    path='test', project=self._TEST_DEVOPS_PROJECT, query_order='folderAscending')
                mock_update_folders.assert_called_once_with(
                    folder=mock_list_folders.return_value[0], path='test', project=self._TEST_DEVOPS_PROJECT)
                self.assertEqual(mock_list_folders.return_value[0].path, 'test123')
                self.assertEqual(mock_list_folders.return_value[0].description, 'hello world updated')
                
    
    def test_folder_update_with_invalid_input(self):
        with self.assertRaises(Exception) as exc:
            response = pipeline_folder_update(
                    path='test', project=self._TEST_DEVOPS_PROJECT, organization=self._TEST_DEVOPS_ORGANIZATION)
        self.assertEqual(str(exc.exception),r'Either --new-path or --new-description should be specified.')                
